# -*-coding:utf-8-*-
u"""blenderのaddonとして認識させる.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
import subprocess
import inspect
import sys
import os

import bpy


from .hardware_render_animations.modules.setting import Setting


bl_info = {
    "name": "Hardware Render Animations",
    "author": "Masayuki Osaka",
    "version": (0, 2, 0),
    "blender": (2, 79, 0),
    "location": "Tool bar > Animation tab > Hardware Render Animations",
    "description": "Perform hardware rendering on multiple files",
    "category": "Animation"
}


class AddMark(bpy.types.Operator):
    u"""フォーカスの対象を限定するためのプロパティを付与."""

    bl_idname = 'object.add_mark_for_hardware_render_animations'
    bl_label = 'Add Mark'
    bl_description = 'Adds a property to selected meshes so that they are '\
                     'in focus when capturing.'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        u"""選択状態のメッシュオブジェクトにプロパティ追加."""
        # プロパティの制限
        limit = {
            'min': 1,
            'max': 1,
            'soft_min': 1,
            'soft_max': 1,
            'default': 1,
            'options': None,
            'description': 'If 1, this mesh is included in targets of focus.'
        }

        # メッシュオブジェクトへプロパティ追加
        for obj in bpy.context.selected_objects:
            if not isinstance(obj.data, bpy.types.Mesh):
                continue
            if Setting.REGISTERD_MARK in obj:
                continue
            obj[Setting.REGISTERD_MARK] = 1

            # 制限追加
            if '_RNA_UI' not in obj:
                obj['_RNA_UI'] = {}
            obj['_RNA_UI'][Setting.REGISTERD_MARK] = limit

        return {'FINISHED'}


class RemoveMark(bpy.types.Operator):
    u"""フォーカスの対象を限定するためのプロパティを削除."""

    bl_idname = 'object.remove_mark_for_hardware_render_animations'
    bl_label = 'Remove Mark'
    bl_description = 'Remove the added property form selected meshes.'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        u"""プロパティ除去."""
        for obj in bpy.context.selected_objects:
            # プロパティ除去
            if Setting.REGISTERD_MARK in obj:
                del obj[Setting.REGISTERD_MARK]

            # _RNA_UIから制限を除去
            if '_RNA_UI' not in obj:
                continue
            if Setting.REGISTERD_MARK in obj['_RNA_UI']:
                del obj['_RNA_UI'][Setting.REGISTERD_MARK]

        return {'FINISHED'}


class SelectMarkedMeshes(bpy.types.Operator):
    u"""フォーカス対象を限定するマークの付いて、その値がTrueのメッシュを選択状態にする."""

    bl_idname = 'object.select_marked_meshes_hardware_render_animations'
    bl_label = 'Select Marked Meshes'
    bl_description = 'Select marked meshes to focus on when capturing.'
    bl_options = {'REGISTER', 'UNDO'}

    def to_be_selected(self, obj):
        u"""選択状態になるべきかを返す.

        プロパティがあり、値がTrueであれば選択状態になるべき

        obj: bpy.types.Object()
        return: bool
        """
        if not isinstance(obj.data, bpy.types.Mesh):
            return False
        if Setting.REGISTERD_MARK not in obj:
            return False
        return True

    def execute(self, context):
        u"""実行部."""
        for obj in bpy.context.scene.objects:
            obj.select = self.to_be_selected(obj)

        return {'FINISHED'}


class OpenGui(bpy.types.Operator):
    u"""サブプロセスで設定用GUIを開く."""

    bl_idname = 'object.open_hardware_rendering_window'
    bl_label = 'Open Main Gui'
    bl_description = 'Open gui in a subprocess for hardware rendering '\
                     'of multiple scenes.'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        u"""実行部."""
        current_dir = os.path.dirname(inspect.stack()[0][1])
        # PySide2を持つPythonパスを取得
        user_setting_path = os.path.join(current_dir, 'user_setting.txt')
        with open(user_setting_path) as fobj:
            python_path = fobj.readlines()[1].strip()
        python_path = python_path.replace('\\\\', '\\').replace('/', os.sep)

        pyfile_execute_gui = os.path.join(
            current_dir,
            'hardware_render_animations',
            'call_gui_independent_process.py'
        )
        current_executable = sys.executable

        # サブプロセスでGUIを起動
        command = '"{}" "{}" --dcc "{}"'.format(
            python_path, pyfile_execute_gui, current_executable
        )
        subprocess.Popen(command, shell=True)

        return {'FINISHED'}


class VIEW3D_PT_tools_hardware_render_animations(bpy.types.Panel):
    u"""Tool Bar > Animation へメニュー追加."""

    bl_space_type = 'VIEW_3D'
    bl_region_type = 'TOOLS'
    bl_category = 'Animation'
    bl_context = 'objectmode'
    bl_label = 'Hardware Render Animations'

    def draw(self, context):
        u"""追加の実行."""
        layout = self.layout
        col = layout.column()

        col.label(text='Add/Remove Mark:')
        row = col.row()
        row.operator(AddMark.bl_idname, text='Add Mark')
        row.operator(RemoveMark.bl_idname, text='Remove Mark')
        col.operator(SelectMarkedMeshes.bl_idname, text='Select Marked Meshes')

        col.separator()

        col.label(text='Open Hardware Render Gui:')
        col.operator(OpenGui.bl_idname, text='Open Gui')


def register():
    u"""アドオンの有効化時に呼ばれる."""
    bpy.utils.register_module(__name__)


def unregister():
    u"""アドオンの無効時に呼ばれる."""
    bpy.utils.unregister_module(__name__)


if __name__ == '__main__':
    register()
