# -*-coding:utf-8-*-
u"""トップモジュール.

blender, Mayaでアニメーションのハードウェアレンダーを行う

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import sys
import os
import json
import importlib

from .modules.vector.calc_vector import CalculateContactWithCircle
from .modules.errorlog import ErrorLog
from .modules.setting import Setting
from .modules.pathes import Pathes
from .modules import factory


def _set_dcc_info():
    u"""Setting.dcc_tool設定."""
    # exeファイルの名前に沿って、DCCツールの種類、バージョンを確定
    executer = os.path.splitext(os.path.basename(sys.executable))[0]
    if executer in ('blender', 'maya'):
        dcc_setter = importlib.import_module(
            '.modules.confirm_dcc.confirm_version_{}'.format(executer),
            Setting.PACKAGE_NAME
        )
        dcc_setter.set_version()
    # DCCツールを経由せずに呼び出した場合、暫定的にblenderとする
    else:
        Setting.dcc_tool = Setting.BLENDER279B


def _main_without_gui():
    u"""guiを介さないで実行.

    PySide2を持たないDCCツールでのレンダリングを、別プロセスで起動して行う場合に呼び出される。
    また、単独実行=シーン上でmain(use_gui=False) として実行された場合にも読みだされる。

    PySide2を介さないでの実行のため、progressbarは無し。

    後者は、Settingを手動で書く必要あり
    """
    # guiで作成した設定があれば読み込み
    if os.path.isfile(Setting.TEMPFILE_PATH):
        with open(Setting.TEMPFILE_PATH) as fobj:
            dic = json.load(fobj)
        Setting.set_by_dict(dic[Setting.SETTING_KEY])

    # 実行
    execute_to_scene = factory.execute_to_scene()
    scene_pathes = Pathes.expand_pathes(*Setting.scenes_pathes)

    for scene in sorted(scene_pathes):
        # キャッシュをクリアして実行
        CalculateContactWithCircle.initialize_cache()
        try:
            execute_to_scene(scene)
        except Exception:
            ErrorLog.add_error(scene)


def main(use_gui=True):
    u"""DCCツールの文脈で実行.

    use_gui bool GUIを使用するか。しない場合はSettingを手で記入する必要あり.
    return 使用したGui()
    """
    # 起動の経緯を基にSetting.dcc_toolを設定
    _set_dcc_info()

    # gui起動
    if use_gui:
        gui = factory.gui()
    else:
        # この場合、設定用jsonを用意するか、Settingを手で記入すること
        CalculateContactWithCircle.initialize_cache()  # キャッシュクリア
        _main_without_gui()
        gui = None
    return gui


def register_tool():
    u"""レンダリング対象メッシュ登録用GUI起動.

    PySide2が組み込みされているDCCで使用する
    """
    gui = factory.gui_register()
    gui.show()
