# -*-coding:utf-8-*-
u"""ローテートオーダーに関するクラスをDCCツールに沿った形で返す.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from .. import factory
from ..setting import Setting
from ..vector import calc_vector


class AbstractFrustumPlane(object):
    u"""上下左右の各視錐平面."""

    def __init__(self, *args, **kwargs):
        u"""視錐平面の作成.

        必要な引数
        ・視錐平面の角度 AOV依存の視錐平面の傾き(左右か上下で値が違う)
        ・カメラの回転 視錐平面の法線をカメラに合わせて回転させる
        """
        raise NotImplementedError

    def set_tip_geometry_point(self, points):
        u"""点群の包めるように点群の端を取得.

        視錐平面の法線と点の内積から、距離(正負込み)を取得して判断
        points: 各DCCタイプの点群
        """
        raise NotImplementedError


class TopFrustumPlane(AbstractFrustumPlane):
    u"""上の視錐平面."""


class BottomFrustumPlane(AbstractFrustumPlane):
    u"""下の視錐平面."""


class LeftFrustumPlane(AbstractFrustumPlane):
    u"""左の視錐平面."""


class RightFrustumPlane(AbstractFrustumPlane):
    u"""右の視錐平面."""


class AbstractFrustumOperator(object):
    u"""視錐平面の制御."""

    def __init__(self, camera_object):
        u"""constructor.

        camera_object: 各DCCタイプのカメラオブジェクト
        """
        self.camera = camera_object

    def _set_safe_area(self, render_aperture_h, render_aperture_v):
        u"""セーフエリアをセット.

        render_aperture_h: float: 限界の水平レンダー範囲
        render_aperture_v: float: 限界の垂直レンダー範囲
        return: tuple(float, float): 補正水平レンダー範囲, 補正垂直レンダー範囲
        """
        margin_coef = 1 - Setting.render_margin_percent * 0.01
        return render_aperture_h * margin_coef, render_aperture_v * margin_coef

    def _get_render_apertrue(self):
        u"""レンダリング用矩形のapertureを取得.

        この辺数とfocal_lengthで視錐平面の角度を取得可能になる。
        レンダリング縦横比、カメラのfilm_aperture, 他設定から算出。

        return: tuple(float, float): 点群を収めるapertureの横、縦
        """
        raise NotImplementedError

    def set_current_frustums(self):
        u"""現在のカメラのアトリビュートから、視錐平面を算出."""
        raise NotImplementedError

    def set_tip_geometry_points(self, points):
        u"""各視錐平面がフィットする点を取得.

        points: フィットさせる点群
        """
        raise NotImplementedError


class AbstractFitCamera(object):
    u"""カメラの設定とジオメトリの点群から、カメラ位置を設定する."""

    def __init__(self, camera_object):
        u"""constructor.

        camera_object: 各DCCタイプのカメラオブジェクト
        """
        self.camera = camera_object
        self.frustum_operator = AbstractFrustumOperator(camera_object)

    @property
    def name(self):
        u"""カメラの名前を返す."""
        raise NotImplementedError

    def hide(self):
        u"""このカメラを非表示に."""
        raise NotImplementedError

    def _get_camera_view_direction(self):
        u"""カメラの視線ベクトルを返す.

        return: Vector3D()
        """
        raise NotImplementedError

    def _get_fit_position(self, top_fit_point, bottom_fit_point,
                          left_fit_point, right_fit_point,
                          top_normal, bottom_normal,
                          left_normal, right_normal):
        u"""各視錐平面の情報とカメラの向きから、カメラの新しい位置を計算.

        args: いずれもVector3D(): 各視錐平面の点と法線
        return: Vector3D()
        """
        # 交線
        line1_p, line1_n = calc_vector.get_cross_line_of_2planes(
            top_fit_point, bottom_fit_point, top_normal, bottom_normal
        )
        line2_p, line2_n = calc_vector.get_cross_line_of_2planes(
            left_fit_point, right_fit_point, left_normal, right_normal
        )
        # 交線の最近接点
        point1, point2 = calc_vector.get_closest_points_of_2lines(
            line1_p, line2_p, line1_n, line2_n
        )
        # 視線の後ろのほうがカメラ位置
        camera_direction = self._get_camera_view_direction()
        dot1 = camera_direction * point1
        dot2 = camera_direction * point2
        if dot1 > dot2:
            return point2
        return point1

    def _recalculate_points_for_turntable(self, frustums, points):
        u"""点群と中心の描く円から、視錐平面の位置算出に必要な点を取得.

        frustums: tuple(各DCCのAbstractFrustumPlaneのサブクラスインスタンス,)
        points: 各DCCタイプの点群
        return: list(tuple(float, float, float),)
        """
        # 各点と中心の距離、高さを取得
        new_points = []
        # 各視錐平面用の接点を取得
        for fructum in frustums:
            new_points += calc_vector.CalculateContactWithCircle.get_contacts(
                fructum.normal, points
            )
        return new_points

    def _calculate_camera_position_from_frustums(self, points):
        u"""視錐平面から、カメラの位置の算出.

        points: 各DCCタイプの点群
        return: Vector3D()
        """
        # カメラ設定から、各視錐平面を初期設定
        self.frustum_operator.set_current_frustums()
        frustums = (
            self.frustum_operator.top_frustum,
            self.frustum_operator.bottom_frustum,
            self.frustum_operator.left_frustum,
            self.frustum_operator.right_frustum
        )
        # 視錐平面の確定に必要な点を再設定
        if Setting.tool_purpose == Setting.ToolPurpose.kTurntable:
            points = self._recalculate_points_for_turntable(frustums, points)
        # 全点群を内側に捉えられる、各視錐平面の位置を取得
        self.frustum_operator.set_tip_geometry_points(points)
        # 各視錐平面の位置から、カメラ位置を算出
        points = [getattr(x, 'tip_geometry_point') for x in frustums]
        normals = [getattr(x, 'normal') for x in frustums]
        points_and_normals = [factory.vector3d(x) for x in points + normals]
        return self._get_fit_position(*points_and_normals)

    def set_current_camera_position(self, points, center=None, height=None):
        u"""視錐平面とジオメトリの端の点を基に、カメラ位置を設定.

        points: 各DCCタイプの点群: ジオメトリ点群
        center: 各DCCタイプの点: 点群の2次元上の中心点(Turntableでのみ有用)
        heights: tuple(float,): 各点の高さ(Turntableでのみ有用)
        """
        raise NotImplementedError

    def set_keyframe(self):
        u"""カメラのカレントtranslate, rotateへキーを打つ."""
        raise NotImplementedError

    def set_turntable_animation(self, center2d):
        u"""ターンテーブルの階層を作成し、階層にキーを打つ.

        center2d: tuple(float, float): 回転中心
        """
        raise NotImplementedError
