# -*-coding:utf-8-*-
u"""blender(Numpy)での、カメラの扱いに関するクラス.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from math import atan, pi

import bpy
import mathutils as mu

from . import camera_abstract as camabc
from .. import factory
from ..setting import Setting


class RotateVector:
    u"""np.array(ベクトル)を、回転行列を使用して回転させるクラス."""

    @classmethod
    def _rx(cls, radians):
        u"""X軸回転行列(4x4)を作成."""
        return mu.Matrix.Rotation(radians, 4, 'X')

    @classmethod
    def _ry(cls, radians):
        u"""Y軸回転行列(4x4)を作成."""
        return mu.Matrix.Rotation(radians, 4, 'Y')

    @classmethod
    def _rz(cls, radians):
        u"""Z軸回転行列(4x4)を作成."""
        return mu.Matrix.Rotation(radians, 4, 'Z')

    @classmethod
    def x_rotate(cls, vector, radians):
        u"""ベクトルをX軸回転.

        vector: mathutils.Vector()
        radians: float()
        return: mathutils.Vector()
        """
        return cls._rx(radians) * vector

    @classmethod
    def y_rotate(cls, vector, radians):
        u"""ベクトルをY軸回転.

        vector: mathutils.Vector()
        radians: float()
        return: mathutils.Vector()
        """
        return cls._ry(radians) * vector

    @classmethod
    def rotate3axes(cls, vector, x, y, z, order):
        u"""ベクトルを3軸回転.

        vector: mathutils.Vector()
        x, y, z: いずれもradians: 回転角度
        order: str(): 'XYZ', 'ZYX', などのいずれか
        return: mathutils.Vector()
        """
        # 回転行列の適用順を決める
        rotator = {'X': cls._rx, 'Y': cls._ry, 'Z': cls._rz}
        rotator1 = rotator[order[0]]
        rotator2 = rotator[order[1]]
        rotator3 = rotator[order[2]]
        angle = {'X': x, 'Y': y, 'Z': z}
        # 各回転順での回転角
        angle1 = angle[order[0]]
        angle2 = angle[order[1]]
        angle3 = angle[order[2]]
        # ベクトル回転
        fixed_vector = (
            rotator3(angle3) *
            rotator2(angle2) *
            rotator1(angle1) *
            vector
        )
        return fixed_vector


class AbstractFrustumPlane(camabc.AbstractFrustumPlane):
    u"""Blenderでの各視錐平面の抽象クラス."""

    BASE_NORMAL = mu.Vector([0, 1, 0])  # aovが0の時の視錐平面の法線
    SLOPE_SIGN = 1  # aov反映時の角度がプラスかマイナスか(1 OR -1)
    ROTATOR = RotateVector.x_rotate

    def __init__(self, slope, camera_rotation):
        u"""constructor.

        slope: radians: 視錐平面の傾き(AOV依存)。上下か左右かで値は変わる
        camera_rotation mathutils.Euler()
        """
        base_normal = self.ROTATOR(self.BASE_NORMAL, slope * self.SLOPE_SIGN)
        # カメラ自身の回転を加える
        self.normal = RotateVector.rotate3axes(
            base_normal,
            camera_rotation.x,
            camera_rotation.y,
            camera_rotation.z,
            camera_rotation.order
        )

    def set_tip_geometry_point(self, points):
        u"""点群の包めるように点群の端を取得.

        視錐平面の法線と点の内積から、距離(正負込み)を取得して判断
        points: list of mathutils.Vector()
        """
        singed_distances = [self.normal.dot(x) for x in points]
        min_signed_distance = min(singed_distances)
        index = singed_distances.index(min_signed_distance)
        self.tip_geometry_point = points[index]


class TopFrustumPlane(AbstractFrustumPlane):
    u"""上の視錐平面."""

    BASE_NORMAL = mu.Vector([0, -1, 0])
    SLOPE_SIGN = 1
    ROTATOR = RotateVector.x_rotate


class BottomFrustumPlane(AbstractFrustumPlane):
    u"""下の視錐平面."""

    BASE_NORMAL = mu.Vector([0, 1, 0])
    SLOPE_SIGN = -1
    ROTATOR = RotateVector.x_rotate


class LeftFrustumPlane(AbstractFrustumPlane):
    u"""左の視錐平面."""

    BASE_NORMAL = mu.Vector([1, 0, 0])
    SLOPE_SIGN = 1
    ROTATOR = RotateVector.y_rotate


class RightFrustumPlane(AbstractFrustumPlane):
    u"""右の視錐平面."""

    BASE_NORMAL = mu.Vector([-1, 0, 0])
    SLOPE_SIGN = -1
    ROTATOR = RotateVector.y_rotate


def _effective_rectangle_fit_horizontal(width, height, sensor_x, **kwargs):
    u"""Horizontalフィットでの、縦横を返す.

    return: tuple(float, float): aov算出に必要な矩形
    """
    return (sensor_x, sensor_x / width * height)


def _effective_rectangle_fit_vertical(width, height, sensor_y, **kwargs):
    u"""Verticalフィットでの、縦横を返す.

    return: tuple(float, float): aov算出に必要な矩形
    """
    return (sensor_y / height * width, sensor_y)


def _effective_rectangle_fit_auto(width, height, sensor_x, **kwargs):
    u"""Autoフィットでの、縦横を返す.

    AutoはsensorX ** 2四方の正方形にフィットする仕様

    return: tuple(float, float): aov算出に必要な矩形
    """
    sensor_size = sensor_x
    if width / height > 1:
        return _effective_rectangle_fit_horizontal(width, height, sensor_size)
    return _effective_rectangle_fit_vertical(width, height, sensor_size)


def effective_rectangle(sensor_fit):
    u"""sensor_fitのモードによって、実効フィルムサイズを計算する関数を返す.

    sensor_fit: str()
    return _efective_rectangle_fit_●●
    """
    use_func = {
        'HORIZONTAL': _effective_rectangle_fit_horizontal,
        'VERTICAL': _effective_rectangle_fit_vertical,
        'AUTO': _effective_rectangle_fit_auto
    }
    return use_func[sensor_fit]


class FrustumOperator(camabc.AbstractFrustumOperator):
    u"""視錐平面の制御."""

    def __init__(self, camera_object):
        u"""constructor.

        camera_object: bpy.data.objectsのカメラ
        """
        self.camera = camera_object
        self.camera_data = camera_object.data

    def _get_render_aperture(self):
        u"""レンダリング用矩形のapertureを取得.

        filmApertureをカメラのsenserサイズ, レンダリング縦横比, 画像比で補正
        return: tuple(float, float): 点群を収めるapertureの横、縦
        """
        # レンダー画像伸縮を逆算
        pixel_aspect_x = bpy.context.scene.render.pixel_aspect_x
        pixel_aspect_y = bpy.context.scene.render.pixel_aspect_y
        render = bpy.context.scene.render
        remodeled_resolution_x = render.resolution_x * pixel_aspect_x
        remodeled_resolution_y = render.resolution_y * pixel_aspect_y
        # レンダー設定から、aov算出に有効なレンダー縦横比を取得
        sensor_x = self.camera_data.sensor_width
        sensor_y = self.camera_data.sensor_height
        func = effective_rectangle(self.camera_data.sensor_fit)
        effective_x, effective_y = func(
            width=remodeled_resolution_x,
            height=remodeled_resolution_y,
            sensor_x=sensor_x,
            sensor_y=sensor_y
        )
        return self._set_safe_area(effective_x, effective_y)

    def set_current_frustums(self):
        u"""現在のカメラのアトリビュートから、視錐平面を算出.

        rotation_euler前提。
        拡張する時はrotation_quaternionなどをrotion_eulerへ変更する必要あり。
        """
        # 各視錐平面の角度算出
        h_aperture, v_aperture = self._get_render_aperture()
        h_slope = atan(h_aperture * 0.5 / self.camera_data.lens)
        v_slope = atan(v_aperture * 0.5 / self.camera_data.lens)
        # カメラの回転を取得
        camera_rotation = self.camera.rotation_euler
        # 各視錐平面を取得
        self.top_frustum = TopFrustumPlane(v_slope, camera_rotation)
        self.bottom_frustum = BottomFrustumPlane(v_slope, camera_rotation)
        self.left_frustum = LeftFrustumPlane(h_slope, camera_rotation)
        self.right_frustum = RightFrustumPlane(h_slope, camera_rotation)

    def set_tip_geometry_points(self, points):
        u"""各視錐平面がフィットする点を取得.

        points: np.array([np.array(), np.array(), ...]): 点群アレイ
        """
        for frustum in (self.top_frustum, self.bottom_frustum,
                        self.left_frustum, self.right_frustum):
            frustum.set_tip_geometry_point(points)


class FitCamera(camabc.AbstractFitCamera):
    u"""カメラの設定とジオメトリの点群から、カメラ位置を設定する."""

    def __init__(self, camera_object):
        u"""constructor.

        camera_object: bpy.data.objectsのカメラ
        """
        self.camera = camera_object.data
        self.frustum_operator = FrustumOperator(camera_object)
        self.transform = camera_object

    @property
    def name(self):
        u"""カメラの名前を返す(bpy.types.objectのほう)."""
        return self.transform.name

    def hide(self):
        u"""このカメラを非表示に."""
        self.transform.hide = True

    def _get_camera_view_direction(self):
        u"""カメラの視線ベクトルを返す.

        return: Vector3D()
        """
        camera_matrix = self.transform.matrix_world
        # 読みやすさのため、column-major order を row major order へ
        transpose = camera_matrix.to_3x3().transposed()
        direction = -mu.Vector(transpose[2])
        return factory.vector3d(direction)

    def set_current_camera_position(self, points):
        u"""視錐平面とジオメトリの端の点を基に、カメラ位置を設定.

        points: list(mathutils.Vector(),)
        """
        # 視錐平面の確定とカメラ位置の算出
        new_position = self._calculate_camera_position_from_frustums(points)
        # カメラ移動
        self.transform.matrix_world.translation = new_position.delegate

    def set_keyframe(self):
        u"""カメラのカレントtranslate, rotateへキーを打つ.

        rotation_euler前提。拡張する場合は、キーを打つ対象を条件分け
        """
        self.transform.keyframe_insert('location')
        self.transform.keyframe_insert('rotation_euler')

    def set_turntable_animation(self, center2d):
        u"""ターンテーブルの階層を作成し、階層にキーを打つ.

        center2d: tuple(float, float): 回転中心
        """
        # 回転土台を作成し、カメラの親にする
        bpy.ops.object.empty_add(location=(center2d[0], center2d[1], 0.0))
        empty = bpy.context.scene.objects.active
        self.transform.parent = empty
        self.transform.matrix_parent_inverse = empty.matrix_world.inverted()

        bpy.ops.object.select_all(action='DESELECT')
        empty.select = True

        # 回転土台にキーを打つ
        bpy.context.scene.frame_set(Setting.start_frame)
        empty.keyframe_insert(data_path='rotation_euler')
        bpy.context.scene.frame_set(Setting.end_frame + 1)
        empty.rotation_euler = mu.Euler(
            (0, 0, pi * 2 * Setting.turntimes_during_animation),
            'XYZ'
        )
        empty.keyframe_insert(data_path='rotation_euler')

        # キーをLINEAR補完にする
        z_cv = empty.animation_data.action.fcurves[-1]
        z_cv.keyframe_points[0].interpolation = 'LINEAR'

        # 回転土台を非表示に(blenderではロケータシェイプが映るため)
        empty.hide = True
