# -*-coding:utf-8-*-
u"""Mayaのメソッドに沿った、カメラに関するクラス.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from math import atan, pi

import maya.api.OpenMaya as om2
import maya.api.OpenMayaAnim as om2a
import maya.cmds as cmds

from . import camera_abstract as camabc
from .. import factory
from ..setting import Setting


class AbstractFrustumPlane(camabc.AbstractFrustumPlane):
    u"""Mayaでの各視錐平面の抽象クラス."""

    BASE_NORMAL = om2.MVector(0, -1, 0)  # aovが0の時の視錐平面の法線
    SLOPE_SIGN = 1  # aov反映時の角度がプラスかマイナスか(1 OR -1)
    SLOPE_AXIS = om2.MVector.kXaxis  # aov反映時の回転軸

    def __init__(self, slope, camera_rotation):
        u"""constructor.

        slope: radians: 視錐平面の傾き(AOV依存)。上下か左右かで値は変わる
        camera_rotation: OpemMaya.MEulerRotation()
        """
        base_normal = self.BASE_NORMAL.rotateBy(self.SLOPE_AXIS,
                                                slope * self.SLOPE_SIGN)
        # カメラ自身の回転を加える
        self.normal = base_normal.rotateBy(camera_rotation)

    def set_tip_geometry_point(self, vectors):
        u"""点群の包めるように点群の端を取得.

        視錐平面の法線と点の内積から、距離(正負込み)を取得して判断
        points: list([om2.MVector(), ...])
        """
        singed_distances = [self.normal * vec for vec in vectors]
        min_signed_distance = min(singed_distances)
        index = singed_distances.index(min_signed_distance)
        self.tip_geometry_point = om2.MPoint(vectors[index])


class TopFrustumPlane(AbstractFrustumPlane):
    u"""上の視錐平面."""

    BASE_NORMAL = om2.MVector(0, -1, 0)
    SLOPE_SIGN = 1
    SLOPE_AXIS = om2.MVector.kXaxis


class BottomFrustumPlane(AbstractFrustumPlane):
    u"""下の視錐平面."""

    BASE_NORMAL = om2.MVector(0, 1, 0)
    SLOPE_SIGN = -1
    SLOPE_AXIS = om2.MVector.kXaxis


class LeftFrustumPlane(AbstractFrustumPlane):
    u"""左の視錐平面."""

    BASE_NORMAL = om2.MVector(1, 0, 0)
    SLOPE_SIGN = 1
    SLOPE_AXIS = om2.MVector.kYaxis


class RightFrustumPlane(AbstractFrustumPlane):
    u"""右の視錐平面."""

    BASE_NORMAL = om2.MVector(-1, 0, 0)
    SLOPE_SIGN = -1
    SLOPE_AXIS = om2.MVector.kYaxis


def get_render_aperture_horizontal(film_aperture_x):
    u"""レンダリングのapertureを取得(カメラfilmFitがhorizontal時).

    return 2 element: 横、縦のレンダリング用aperture
    """
    fixed_aperture_y = (film_aperture_x * Setting.resolution_y /
                        Setting.resolution_x)
    return film_aperture_x, fixed_aperture_y


def get_render_aperture_vertical(film_aperture_y):
    u"""レンダリングのapertureを取得(カメラfilmFitがvertical時).

    return 2 element: 横、縦のレンダリング用aperture
    """
    fixed_aperture_x = (film_aperture_y * Setting.resolution_x /
                        Setting.resolution_y)
    return fixed_aperture_x, film_aperture_y


class FrustumOperator(camabc.AbstractFrustumOperator):
    u"""視錐平面の制御."""

    def __init__(self, camera_object):
        u"""constructor.

        camera_object: om2.MFnCamera()
        """
        self.camera = camera_object

    def _get_render_aperture(self):
        u"""レンダリング用矩形のpaertureを取得.

        filmApertureをカメラのfilmFit, レンダリング縦横から補正算出
        return: tuple(float, float): 点群を収めるapertureの横、縦
        """
        # フィルムとレンダリングの、アパーチャと比率を取得
        film_aperture_x = self.camera.horizontalFilmAperture * 25.4
        film_aperture_y = self.camera.verticalFilmAperture * 25.4
        film_aspect_ratio = film_aperture_x / film_aperture_y
        render_aspect_ratio = Setting.resolution_x / Setting.resolution_y
        # カメラのfilmFit依存のレンダリングアパーチャを取得
        render_aperture_h = get_render_aperture_horizontal(film_aperture_x)
        render_aperture_v = get_render_aperture_vertical(film_aperture_y)
        # カメラのfilmFit次第でレンダリング用視錐平面が変わる
        if self.camera.filmFit == om2.MFnCamera.kHorizontalFilmFit:
            return self._set_safe_area(*render_aperture_h)
        if self.camera.filmFit == om2.MFnCamera.kVerticalFilmFit:
            return self._set_safe_area(*render_aperture_v)
        if self.camera.filmFit == om2.MFnCamera.kFillFilmFit:
            if film_aspect_ratio > render_aspect_ratio:
                return self._set_safe_area(*render_aperture_v)
            return self._set_safe_area(*render_aperture_h)
        if self.camera.filmFit == om2.MFnCamera.kOverscanFilmFit:
            if film_aspect_ratio > render_aspect_ratio:
                return self._set_safe_area(*render_aperture_h)
            return self._set_safe_area(*render_aperture_v)

    def set_current_frustums(self):
        u"""現在のカメラのアトリビュートから、視錐平面を算出."""
        # 各視錐平面の角度算出
        f_len = self.camera.focalLength
        h_aperture, v_aperture = self._get_render_aperture()
        h_slope = atan(h_aperture * 0.5 / f_len)
        v_slope = atan(v_aperture * 0.5 / f_len)
        # カメラの回転を取得
        camera_rotation = om2.MFnTransform(self.camera.parent(0)).rotation()
        # 各視錐平面を取得
        self.top_frustum = TopFrustumPlane(v_slope, camera_rotation)
        self.bottom_frustum = BottomFrustumPlane(v_slope, camera_rotation)
        self.left_frustum = LeftFrustumPlane(h_slope, camera_rotation)
        self.right_frustum = RightFrustumPlane(h_slope, camera_rotation)

    def set_tip_geometry_points(self, points):
        u"""各視錐平面がフィットする点を取得.

        points: フィットさせる点群
        """
        vectors = [om2.MVector(x) for x in points]
        for frustum in (self.top_frustum, self.bottom_frustum,
                        self.left_frustum, self.right_frustum):
            frustum.set_tip_geometry_point(vectors)


class FitCamera(camabc.AbstractFitCamera):
    u"""カメラの設定とジオメトリの点群から、カメラ位置を設定する."""

    def __init__(self, camera_object):
        u"""constructor.

        camera_object: om2.MFnCamera()
        """
        self.camera = camera_object
        self.frustum_operator = FrustumOperator(camera_object)
        self.transform = om2.MFnTransform(self.camera.parent(0))

    @property
    def name(self):
        u"""カメラの名前を返す."""
        return self.camera.name()

    def hide(self):
        u"""このカメラを非表示に."""
        cmds.setAttr(self.name + '.v', 0)

    def _get_camera_view_direction(self):
        u"""カメラの視線ベクトルを返す.

        return: Vector3D()
        """
        direction = self.camera.viewDirection(om2.MSpace.kWorld)
        return factory.vector3d(direction)

    def set_current_camera_position(self, points):
        u"""視錐平面とジオメトリの端の点を基に、カメラ位置を設定.

        points: om2.MPointArray()
        """
        # 視錐平面の確定とカメラ位置の算出
        new_position = self._calculate_camera_position_from_frustums(points)
        # カメラ移動
        cmds.xform(self.transform.name(), t=new_position.delegate, ws=1, a=1)

    def set_keyframe(self):
        u"""カメラのカレントtranslate, rotateへキーを打つ."""
        kwargs = dict(ott='linear', itt='linear')
        cmds.setKeyframe(self.transform.name(), at='t', **kwargs)
        cmds.setKeyframe(self.transform.name(), at='r', **kwargs)

    def set_turntable_animation(self, center2d):
        u"""ターンテーブルの階層を作成し、階層にキーを打つ.

        center2d: tuple(float, float): 回転中心
        """
        # 回転土台(カメラの親)の作成
        center = om2.MVector(center2d[0], 0.0, center2d[1])
        rotate_base = self._create_parent_of_camera(center)
        # 回転土台にキーを打つ
        anim = om2a.MFnAnimCurve()
        anim.create(anim.kAnimCurveTA)
        unit = om2.MTime.uiUnit()
        start = om2.MTime(Setting.start_frame, unit)
        end = om2.MTime(Setting.end_frame, unit) + 1
        anim.addKeys(
            (start, end),
            (0, pi * 2 * Setting.turntimes_during_animation),
            anim.kTangentLinear, anim.kTangentLinear
        )
        plug_anm = anim.findPlug('output', 0)
        plug_trs = rotate_base.findPlug('ry', 0)
        om2.MDGModifier().connect(plug_anm, plug_trs).doIt()

    def _create_parent_of_camera(self, center):
        u"""カメラの親として、回転させる土台のノードを作成.

        center: om2.MVector()
        return: om2.MFnTransform()
        """
        parent = om2.MFnTransform()
        parent.create()
        parent.setTranslation(center, om2.MSpace.kTransform)
        cmds.makeIdentity(parent.partialPathName(), a=1, t=1)
        parent.addChild(self.transform.object())
        return parent
