# -*-coding:utf-8-*-
u"""DCCツールに応じたインスタンスを提供する.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import importlib

from .setting import Setting


def _get_module(module_prefix):
    u"""DCCツールに応じたモジュールを提供.

    module_prefix: DCCツールによらず共通のモジュール名部分
    e.g. "camera"
      -> import_module('.modules.camera.camera_maya2018', PACKAGE)

    return: module
    """
    module = importlib.import_module(
        '.modules.{0}.{0}_{1}'.format(module_prefix, Setting.dcc_tool),
        Setting.PACKAGE_NAME
    )
    return module


def confirm_setting():
    u"""DCCツールに応じたConfirmSettingインスタンスを提供する.

    return: ConfirmSetting()
    """
    module = _get_module('confirm_setting')
    return module.ConfirmSetting()


def fit_camera(camera_object):
    u"""DCCツールに応じたFitCameraインスタンスを提供する.

    camera_object: 各DCCタイプのカメラオブジェクト
    return: FitCamera()
    """
    module = _get_module('camera')
    return module.FitCamera(camera_object)


def vector3d(vector_object):
    u"""DCCツールに適したVector3Dインスタンスを提供.

    vector_object: 各DCCタイプのVectorオブジェクト
    return: Vector3D()
    """
    module = _get_module('vector')
    return module.Vector3D(vector_object)


def point_container():
    u"""DCCツールに適したPointContainerインスタンスを提供.

    return: PointContainer()
    """
    module = _get_module('mesh_points')
    return module.PointContainer()


def renderer(point_container, *fit_cameras):
    u"""DCCツールに適したRendererインスタンスを提供.

    point_container: PointContainer()
    *fit_cameras: FitCamera()
    return: Renderer()
    """
    module = _get_module('renderer')
    return module.Renderer(point_container, *fit_cameras)


def execute_to_scene():
    u"""DCCツールに適した実行ファンクションを提供.

    return: function: main_operator.main_operator_xxx.main
    """
    module = _get_module('main_func')
    return module.execute_to_scene


def gui():
    u"""DCCツールに適したGuiインスタンスを提供.

    return: Gui()
    """
    module = _get_module('gui')
    return module.Gui()


def gui_register():
    u"""DCCツールに適したRegisterGuiインスタンスを提供.

    return: RggisterGui()
    """
    module = _get_module('gui_register')
    return module.RegisterGui()
