# -*-coding:utf-8-*-
u"""guiのトップ.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import subprocess
import webbrowser
import tempfile
import inspect

import sys
import os

from PySide2.QtWidgets import (
    QMainWindow, QListWidget, QListWidgetItem, QAbstractItemView, QMenu,
    QMessageBox, QComboBox, QFileDialog, QStyle, QColorDialog, QStyleFactory,
    QPushButton, QHBoxLayout, QProgressBar
)
from PySide2.QtGui import (
    QBrush, QPalette, QIntValidator, QDoubleValidator, QColor
)
from PySide2.QtCore import Qt

from .import_export_mixin import ImportExportMixin, ColorWidget
from .resources import ui_main
from ..errorlog import ErrorLog
from ..setting import Setting
from ..pathes import Pathes
from ..vector.calc_vector import CalculateContactWithCircle
from .. import factory


class ProgressBar(QProgressBar):
    u"""progressbar."""

    def __init__(self, parent=None, max=max):
        u"""constructor."""
        super(ProgressBar, self).__init__(parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.value = 0
        self.setMinimum(0)
        self.setMaximum(max)
        self.show()

    def step(self):
        u"""progressは1固定."""
        self.value += 1
        self.setValue(self.value)


class ShowMessagebox(object):
    u"""メッセージボックスを出すための共通インターフェース."""

    @classmethod
    def critical(cls, parent=None, text=None, title=Setting.PACKAGE_NAME):
        u"""失敗を伝えるメッセージボックス表示."""
        if text is None:
            exts = Setting.scene_exts()
            text = 'Scene files not found. ("{}")'.format('", "'.join(exts))
        QMessageBox.critical(parent, title, text)

    @classmethod
    def success(cls, parent=None, text=None, title=Setting.PACKAGE_NAME):
        u"""通常終了時の表示."""
        box = QMessageBox()
        box.setWindowTitle(Setting.PACKAGE_NAME)
        box.setIcon(QMessageBox.Information)
        box.setText('Finished All.')
        box.exec_()


class ColorDialog(QColorDialog):
    u"""カラーダイアログ."""

    def __init__(self, parent=None):
        u"""constructor."""
        super(ColorDialog, self).__init__(parent)
        self.rejected.connect(self.rejected_slot)

    def rejected_slot(self):
        u"""escape, cancel時はなにもしない."""
        pass


class ColorButtonWidget(ColorWidget):
    u"""カラーダイアログを開くボタン."""

    def __init__(self, parent=None):
        u"""constructor."""
        super(ColorButtonWidget, self).__init__(parent)

        # ボタン付きレイアウト
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.bt = QPushButton()
        self.bt.setAutoFillBackground(True)
        layout.addWidget(self.bt, 1, Qt.AlignCenter)
        self.set_value()

        self.bt.setStyle(QStyleFactory.create('Windows'))

        # カラーダイアログ
        self.dialog = QColorDialog()
        self.dialog.accepted.connect(self._dialog_accepted_slot)

        self.bt.clicked.connect(self._open_colordialog)

    def set_value(self, rgb=None):
        u"""Settingの背景色を反映.

        rgb: tuple(float(),)
        """
        if rgb is None:
            rgb = Setting.bg_color
        qcol = QColor()
        qcol.setRgbF(*rgb)
        palette = QPalette()
        palette.setColor(QPalette.Button, qcol.toRgb())
        self.bt.setPalette(palette)

    def get_value(self):
        u"""色をtupleで取得."""
        col = self.bt.palette().color(QPalette.Button).toRgb()
        return col.redF(), col.greenF(), col.blueF()

    def _open_colordialog(self):
        current_rgb = self.get_value()
        col = QColor()
        col.setRgbF(*current_rgb)
        self.dialog.setCurrentColor(col)
        self.dialog.exec_()

    def _dialog_accepted_slot(self):
        u"""カラーダイアログによる色を反映."""
        col = self.dialog.currentColor()
        self.set_value(rgb=(col.redF(), col.greenF(), col.blueF()))


class VoidListWidgetItem(QListWidgetItem):
    u"""リストが空の時に、D&Dを促すメッセージを出すためのListWidgetItem."""

    def __init__(self, parent=None):
        u"""constructor."""
        super(VoidListWidgetItem, self).__init__(parent)
        self.setForeground(QBrush(Qt.gray))
        self.setFlags(Qt.NoItemFlags)


class UrlDropListWidget(QListWidget):
    u"""ファイル、フォルダをドロップ可能なQListWidget."""

    def __init__(self, parent=None):
        u"""constructor."""
        super(UrlDropListWidget, self).__init__(parent)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setAcceptDrops(True)
        # 右クリックメニュー
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_rmouse_menu)

    def dragEnterEvent(self, event):
        u"""virtual."""
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        u"""virtual."""
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        u"""virtual."""
        if not event.mimeData().hasUrls():
            event.ignore()
            return
        pathes = [x.toLocalFile() for x in event.mimeData().urls()]
        self.add_pathes(pathes)

    def _add_blank_items(self):
        u"""リストが空の時に入れる案内."""
        exts = Setting.scene_exts()
        blank_msgs = (
            '',
            '',
            '  Drop to add files("{}") or folders.'.format('", "'.join(exts)),
            '',
            '  Right mouse to delete or expand selected items.'
        )
        for msg in blank_msgs:
            item = VoidListWidgetItem()
            item.setText(msg)
            self.addItem(item)

    def _show_rmouse_menu(self, mouse_position):
        u"""右マウスでのメニューを表示."""
        items = self.selectedItems()
        if not items:
            return
        menu = QMenu(self)
        delete = menu.addAction('Delete Selected Path')
        delete.triggered.connect(lambda: self._delete_pathes(None))
        # フォルダ選択があれば展開メニューを追加
        for item in items:
            if os.path.isdir(item.text()):
                expand = menu.addAction('Expand Folder')
                expand.triggered.connect(self._expand_folders)
                break
        menu.exec_(self.mapToGlobal(mouse_position))

    def add_pathes(self, pathes, force_add_invalid_pathes=False):
        u"""アイテムを追加.

        pathes: [unicode,]: 追加パスリスト
        force_add_invalid_pathes: bool:
            存在しないパス、ツールに合わないものでも追加するか。（グレーアウトはされる）
        """
        # 空の時用のダミーアイテムを捨てる
        if isinstance(self.item(0), VoidListWidgetItem):
            self.clear()

        if not force_add_invalid_pathes:
            # フォルダと、拡張子がDCCツールに合うもののみ入れる
            exts = Setting.scene_exts()
            pathes = [
                x for x in pathes if os.path.isdir(x) or
                any(x.endswith(ext) for ext in exts)
            ]

        # 追加とソート
        style = self.style()
        exist_pathes = [self.item(i).text() for i in range(self.count())]
        new_pathes = [x for x in pathes if x not in exist_pathes]
        for path in new_pathes:
            if os.path.isfile(path):
                icon = style.standardIcon(style.SP_FileIcon)
            else:
                icon = style.standardIcon(style.SP_DirIcon)
            self.addItem(QListWidgetItem(icon, path))
        self.sortItems(Qt.AscendingOrder)

        # 結果的にリストが空になっていれば、案内文を入れる
        if self.count() == 0:
            self._add_blank_items()

        # 実在しないものと、実在しても拡張子がDCCツールに合わないファイルは警告の色を入れる
        exts = Setting.scene_exts()
        for i in range(self.count()):
            item = self.item(i)
            if isinstance(item, VoidListWidgetItem):
                continue
            path = item.text()
            if os.path.isdir(path):
                continue
            if os.path.isfile(path):
                if any(path.endswith(x) for x in exts):
                    continue
            # 不在確定のアイテムに色を付ける
            item.setForeground(QBrush(Qt.gray))

    def _delete_pathes(self, specific_items=None):
        u"""選択中のアイテムを消す.

        specific_items: [listWidgetItem(),]: 明示して削除させるパスリスト
            Noneの場合、選択されているものを削除
        """
        if specific_items is None:
            items = self.selectedItems()
        else:
            items = specific_items

        for item in items:
            self.takeItem(self.row(item))
        if self.count() == 0:
            self._add_blank_items()

    def _expand_folders(self):
        u"""選択中のフォルダを展開して、中のDCCツールに合うファイルをリストへ入れる."""
        dirs = [x for x in self.selectedItems() if os.path.isdir(x.text())]

        scenes = Pathes.expand_pathes(*[x.text() for x in dirs])
        self.add_pathes(scenes)
        self._delete_pathes(dirs)
        self.clearSelection()

    def get_raw_pathes(self):
        u"""記載のリストををのまま返す.

        return: list[unicode,]
        """
        return [self.item(i).text() for i in range(self.count())]

    def get_pathes(self):
        u"""DCCツールの処理対象になるパス一覧を取得.

        return: list[unicode,]
        """
        path_texts = self.get_raw_pathes()
        files = Pathes.expand_pathes(*path_texts)
        if not files:
            ShowMessagebox.critical(self)
            return
        return files


class AbstractRenderFormatComboBox(QComboBox):
    u"""保存時のフォーマットを決めるcombobox.

    DCCツールによって、自身と、関連Widgetの表示内容を変える
    """

    class AbstractFormat(object):
        u"""フォーマットを定義する基底クラス."""

        sub_labels = ('',)
        display_encode = True
        display_quality = True

    def __init__(self, encode_combobox, quality_frame, parent=None):
        u"""constructor.

        encode_combobox: QComboBox()
        quality_form: QForm()
        """
        super(AbstractRenderFormatComboBox, self).__init__(parent)
        self.encode_cmb = encode_combobox
        self.quality_frame = quality_frame

        # 継承先から、拡張子とサブ要素の情報をクラスで取得
        self.ext_classes = [
            x for x in self.__class__.__dict__.values() if
            isinstance(x, type) and issubclass(x, self.AbstractFormat)
        ]
        # 自身の項目を埋める
        self.ext_classes.sort(key=lambda x: x.__name__)
        for ext_class in self.ext_classes:
            self.addItem(ext_class.__name__.lower())

        # encode, qualityに関するサブ要素へのシグナル
        self.currentIndexChanged[str].connect(self._set_subcombobox)

    def _set_subcombobox(self, text):
        u"""拡張子のSettingへの反映と、encode, qualityへの表示内容切り替え."""
        # 選択された項目に該当する、サブcombobox, サブform操作用クラスを取得
        for ext_class in self.ext_classes:
            if ext_class.__name__.lower() == text.lower():
                break
        else:
            raise Exception

        # encode設定用サブcomboboxを操作
        self.encode_cmb.parentWidget().setVisible(ext_class.display_encode)
        self.encode_cmb.clear()
        for label in ext_class.sub_labels:
            self.encode_cmb.addItem(label)
        # quality設定用サブframeを操作
        self.quality_frame.setVisible(ext_class.display_quality)

        # Settingへの拡張子反映
        Setting.save_ext = ext_class.__name__.lower()
        movie_exts = ('avi', 'mov', 'mp4')
        if Setting.save_ext in movie_exts:
            Setting.save_as_images = False
        else:
            Setting.save_as_images = True


class AbstractGui(QMainWindow, ui_main.Ui_MainWindow, ImportExportMixin):
    u"""gui."""

    WINDOW_NAME = 'hardware_render_animations_window'

    def __init__(self, parent=None):
        u"""gui."""
        # import, Exportで使用したディレクトリがあれば使いまわすためのパス
        self.ever_handled_setting_path_exist = None

        self._delete_old_gui()
        super(AbstractGui, self).__init__(parent)
        self.setupUi(self)
        self.setObjectName(self.WINDOW_NAME)
        self.setWindowTitle(Setting.PACKAGE_NAME)
        self.setAttribute(Qt.WA_DeleteOnClose)
        # 独自Widgetの挿入
        self.listwidget = UrlDropListWidget()
        self.layout_list.insertWidget(0, self.listwidget)
        self.cmb_container = self._make_render_format_combobox()
        self.layout_ext.addWidget(self.cmb_container)
        self.bt_color = ColorButtonWidget(self)
        self.render_grid.addWidget(self.bt_color, 3, 1, 1, 2)
        # GUIその他設定
        self._set_validations()
        self._set_init_camera_values()

        self.bt_save_folder.setIcon(
            self.style().standardIcon(QStyle.SP_DialogOpenButton)
        )

        # シグナル
        self.cmb_container.currentIndexChanged[str].connect(
            self._set_background_widget
        )
        self.cmb_encode.currentIndexChanged[str].connect(self._set_encode)
        self.line_quality.textChanged.connect(self._set_quality_slider)
        self.slide_quality.valueChanged.connect(self._set_quality_label)
        self.cb_mask.stateChanged[int].connect(self.bt_color.setHidden)
        self.bt_save_folder.clicked.connect(self._set_absolute_save_path)
        self.bt_render.clicked.connect(self._do_render)
        self.act_import.triggered.connect(self._import_setting)
        self.act_export.triggered.connect(self._export_setting)
        self.show_help.triggered.connect(
            lambda: webbrowser.open_new_tab(Setting.HELP_URL)
        )

        # DCCツール毎に独自なGUI編集
        self._dcc_own_gui_setup()

        # 初期状態の表示を確定
        self._visibility_setup()

        # 余計にGUIサイズが大きくならないようにして表示開始
        self.resize(self.minimumSizeHint())
        self.show()

    def _make_render_format_combobox(self):
        u"""拡張子と関連要素のGUI表示を設定するcomboboxの追加.

        スコープの関係でオーバーライド先で行う
        return: QComboBox()
        """
        raise NotImplementedError

    def _visibility_setup(self):
        u"""初期状態."""
        # リストの操作説明を表示
        self.listwidget._add_blank_items()
        # UI変更に関するスロットを一度起動させ、初期化する
        self.rb_turn.setChecked(True)
        self.rb_animation.setChecked(True)
        self.rb_path_absolute.setChecked(True)
        self.rb_path_relative.setChecked(True)
        self.rb_cam_scene.setChecked(True)
        self.rb_cam_new.setChecked(True)
        self.cmb_container.setCurrentIndex(1)
        self.cmb_container.setCurrentIndex(0)
        # タブをリストタブへ
        self.tabwidget.setCurrentIndex(0)
        # 保存先絶対パス
        self.line_absolute.setText(tempfile.gettempdir())

    def _dcc_own_gui_setup(self):
        u"""DCCツール毎に独自なGUI編集."""
        raise NotImplementedError

    def _delete_old_gui(self):
        u"""既存のUI削除."""
        raise NotImplementedError

    def _set_validations(self):
        u"""入力型制限."""
        validate_dic = dict()
        validate_dic[QDoubleValidator()] = (
            'line_focal_length', 'line_rotx', 'line_roty', 'line_rotz',
            'line_near', 'line_far'
        )
        validate_dic[QIntValidator()] = (
            'line_resolutionx', 'line_resolutiony', 'line_quality',
            'line_anim_start', 'line_anim_end', 'line_turn_times',
            'line_safe_area',
        )
        # 各widgetへ制限を付与
        for validator, widget_names in validate_dic.items():
            for widget_name in widget_names:
                getattr(self, widget_name).setValidator(validator)

    def _camera_default_rotation(self):
        u"""カメラの初期値.

        return: tuple(float, float, float)
        """
        raise NotImplementedError

    def _set_init_camera_values(self):
        u"""カメラ回転、クリッピングの初期値."""
        xyz = self._camera_default_rotation()
        self.line_rotx.setText(str(float(xyz[0])))
        self.line_roty.setText(str(float(xyz[1])))
        self.line_rotz.setText(str(float(xyz[2])))
        near, far = self._camera_default_clippingplane()
        self.line_near.setText(str(float(near)))
        self.line_far.setText(str(float(far)))

    def _set_encode(self, label):
        u"""エンコード設定のスロット."""
        Setting.save_format = label

    def _set_quality_slider(self, value):
        u"""レンダー品質のスライダ値変更."""
        value = int(value)
        self.slide_quality.setValue(value)
        Setting.quality = value

    def _set_quality_label(self, value):
        u"""レンダー品質のlineEdit値変更."""
        str_v = str(value)
        self.line_quality.setText(str_v)
        Setting.quality = value

    def _send_setting(self):
        u"""gui依存の設定をSettingへ送る.

        Setting.scenes_pathesに関しては、設定export時とレンダー時で内容が変わるので別で行う
        (export時: guiそのままを書き出す, render時: 内容を展開して拡張子毎に抜き出す)
        """
        if self.rb_animation.isChecked():
            Setting.tool_purpose = Setting.ToolPurpose.kPlayblast
        else:
            Setting.tool_purpose = Setting.ToolPurpose.kTurntable

        # Capture Settings
        if self.rb_match_perfect.isChecked():
            Setting.camera_name_match_type = Setting.MatchType.kPerfectMatch
        else:
            Setting.camera_name_match_type = Setting.MatchType.kPartialMatch

        if self.rb_anim_whole.isChecked():
            Setting.camera_fit_mode = Setting.FitType.kStay
        else:
            Setting.camera_fit_mode = Setting.FitType.kEachFrameFigure

        if self.rb_turn_whole.isChecked():
            Setting.camera_fit_mode_turntable = Setting.TurnTarget.kAnim
        else:
            Setting.camera_fit_mode_turntable = Setting.TurnTarget.kStay

        if self.rb_turn_origin.isChecked():
            Setting.turntable_center = Setting.TurnCenter.kOrigin
        else:
            Setting.turntable_center = Setting.TurnCenter.kCenterGeometry

        shading = 'k' + self.cmb_shade.currentText()
        Setting.shading = getattr(Setting.ShadingType, shading)

        words = [x.strip() for x in self.line_cam_filter.text().split(',')]
        words += [x.strip() for x in self.line_cam_filter.text().split(' ')]
        Setting.camera_filter_kwords = [x for x in sorted(set(words)) if x]

        Setting.capture_only_register_mesh = self.rb_mesh_regist.isChecked()
        Setting.prefer_in_scene_camera = self.rb_cam_scene.isChecked()
        Setting.camera_focal_length = float(self.line_focal_length.text())
        Setting.camera_near = float(self.line_near.text())
        Setting.camera_far = float(self.line_far.text())
        Setting.cam_rotx = float(self.line_rotx.text())
        Setting.cam_roty = float(self.line_roty.text())
        Setting.cam_rotz = float(self.line_rotz.text())
        Setting.turntimes_during_animation = int(self.line_turn_times.text())

        # Output Settings
        Setting.show_ornament = not(self.cb_hide_ornament.isChecked())
        Setting.prefer_scene_range = (
            self.cb_animrange_from_scene.isChecked()
        )
        Setting.prefer_scene_resolution = (
            self.cb_resolution_from_scene.isChecked()
        )
        Setting.resolution_x = int(self.line_resolutionx.text())
        Setting.resolution_y = int(self.line_resolutiony.text())
        Setting.render_margin_percent = float(self.line_safe_area.text())
        Setting.start_frame = int(self.line_anim_start.text())
        Setting.end_frame = int(self.line_anim_end.text())
        Setting.bg_color = self.bt_color.get_value()
        Setting.bg_mask = self.cb_mask.isChecked()
        Setting.save_ext = self.cmb_container.currentText()
        Setting.save_format = self.cmb_encode.currentText()
        Setting.quality = self.slide_quality.value()

        if self.rb_path_relative.isChecked():
            Setting.specific_save_root_dir = ''
            Setting.relative_save_dir = self.line_relative.text()
        else:
            Setting.specific_save_root_dir = self.line_absolute.text()

    def _set_absolute_save_path(self):
        u"""保存先を絶対パスにするときのフォルダ指定."""
        specified_dir = self.line_absolute.text()
        save_dir = QFileDialog.getExistingDirectory(
            self, 'Save Directory', specified_dir, QFileDialog.ReadOnly
        )
        if save_dir:
            self.line_absolute.setText(save_dir)

    def _set_background_widget(self, ext):
        u"""背景色のwidget回りの表示状態を編集.

        動画では、背景色の指定を必須とする
        ext: unicode: コンボボックスにより設定された拡張子文字列
        """
        self.lbl_bg.setVisible(True)

        if ext in ('avi', 'mp4', 'mov'):
            self.cb_mask.setVisible(False)
            self.bt_color.setVisible(True)

            return

        if 'maya' in Setting.dcc_tool:
            self.cb_mask.setVisible(True)
            if self.cb_mask.isChecked():
                self.bt_color.setVisible(False)
            else:
                self.bt_color.setVisible(True)
        # blenderかつ画像形式では、背景は強制で透明
        else:
            self.cb_mask.setVisible(False)
            self.bt_color.setVisible(False)
            self.lbl_bg.setVisible(False)

    def _do_render(self):
        u"""レンダリング実行スロット."""
        # Settingの項目設定
        self._send_setting()
        # ファイル毎に実行
        scene_pathes = self.listwidget.get_pathes()
        if not scene_pathes:
            return
        execute_to_scene = factory.execute_to_scene()
        progress = ProgressBar(self, len(scene_pathes))
        for scene in sorted(scene_pathes):
            # 半径計算のキャッシュをクリアして実行
            CalculateContactWithCircle.initialize_cache()
            execute_to_scene(scene)
            progress.step()

        # 最後は新規シーンを開いた状態で終了
        execute_to_scene(scene_path=None)
        progress.close()

        ShowMessagebox.success(self)


class AbstractGuiIndependentProcess(AbstractGui):
    u"""DCCツールがPySide2をもっていない場合用の、別プロセスで起動させるGUI.

    レンダリングはまた別プロセスで行う
    """

    def _do_render(self):
        u"""別プロセスでDCCツールを立ち上げ、ハードウェアレンダー実行."""
        # ログファイルの準備
        try:
            ErrorLog.initialize()
        except OSError:
            ShowMessagebox.critical(self, 'Close Log File.')
            return

        # 一時ファイルへ設定情報を書き出し
        scene_pathes = self._save_gui_setting_to_json(Setting.TEMPFILE_PATH)
        if not scene_pathes:
            ShowMessagebox.critical(self)
            return

        # サブプロセスでのDCCツール起動時に、実行させるpythonファイル
        sub_execute_pyfile = os.path.join(
            inspect.stack()[0][1], '../../../render_with_subprocess.py'
        )
        # subprocess用コマンド生成
        command = self._make_subprocess_command(
            os.path.normpath(sub_execute_pyfile)
        )
        # 別プロセスでDCCを立ち上げ、render_with_subprocess.pyによるレンダリング
        proc = subprocess.Popen(
            command, shell=True, stdout=sys.stdout, stderr=sys.stderr
        )
        proc.wait()

        # 終了後、エラーがあれば、ログを表示
        error_message = ErrorLog.get_error_messages()
        if not error_message:
            ShowMessagebox.success(self)
        else:
            subprocess.Popen('call {}'.format(Setting.LOG_PATH), shell=True)
            ShowMessagebox.critical(
                self,
                'Finished, but Error in some scenes.\n\nLook in the\n' +
                Setting.LOG_PATH.replace(os.sep, '/')
            )
            ErrorLog.delete()

    def _make_subprocess_command(self, sub_execute_pyfile, *args):
        u"""別プロセスでDCCツールを立ち上げるためのコマンド作成.

        sub_execute_pyfile: unicode: 別プロセスでDCCを立ち上げる時の引数になる.py
        return: unicode: subprocess でDCCの起動に使う
        """
        raise NotImplementedError

    def _import_and_bg_widget_setting(self):
        u"""jsonからguiへの反映と、特定widgetの表示確定."""
        self._import_setting()
        self._set_background_widget(self.cmb_container.currentText())
