# -*-coding:utf-8-*-
u"""export, importを担当.

gui_abstractが長くなりすぎたので切り分け

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
from collections import OrderedDict
from numbers import Number
import os
import sys
import json
import numbers

from PySide2.QtWidgets import QListWidget, QRadioButton, QComboBox, QWidget
from PySide2.QtWidgets import QLineEdit, QCheckBox, QFileDialog, QMessageBox

from ..setting import Setting


class ColorWidget(QWidget):
    u"""色を扱うWidget用の基底クラス."""


class ImportExportMixin(object):
    u"""export, importを担当する、ミックスインクラス."""

    # jsonで設定をimport, exportする際のキー
    TOOL_NAME_KEY = '_tool_name'
    SETTING_KEY = 'setting'
    GUI_KEY = 'gui'

    @staticmethod
    def texts_in_customlist(customlist):
        u"""QListWidgetのカスタムの、中のテキストを取得.

        customlist QListWidgetカスタムまたはその継承したインスタンス
        return [unicode,]
        """
        texts = [customlist.item(i).text() for i in range(customlist.count())]
        if not texts or not texts[0]:
            return []
        return texts

    # jsonでimport, exportする必要のあるwidgetのベースクラスと、そのget, set名
    WIDGET_ATTRS = (
        (QListWidget, 'get_raw_pathes', 'add_pathes'),
        (QRadioButton, 'isChecked', 'setChecked'),
        (QComboBox, 'currentText', 'setCurrentText'),
        (QLineEdit, 'text', 'setText'),
        (QCheckBox, 'isChecked', 'setChecked'),
        (ColorWidget, 'get_value', 'set_value')
    )
    # import, exportする必要のあるwidgetのベースクラス (QListWidget, ...)
    BASETYPES_WRITE_OUT_WIDGET = tuple([x[0] for x in WIDGET_ATTRS])
    # widgetクラスと、その値を取得するメソッド名を辞書化 {QLineEdit: 'isChecked', ...}
    ATTRNAME_GET_BY_WIDGET = {x[0]: x[1] for x in WIDGET_ATTRS}
    # widgetクラスと、その値を取得するメソッド名を辞書化 {QLineEdit: 'setChecked', ...}
    ATTRNAME_SET_BY_WIDGET = {x[0]: x[2] for x in WIDGET_ATTRS}

    def _set_widget_list_value(self, widget, values):
        u"""セットする値がシーケンスのWidgetへ値を設定.

        widget ヴィジェットインスタンス
        values 設定するイテレータ
        """
        if all(isinstance(x, numbers.Number) for x in values):
            widget.set_value(values)
            return

        # listwidgetの場合は特別
        pathes = values
        if len(values) > 0 and not values[0].strip():
            pathes = []
        self.listwidget.add_pathes(
            pathes, force_add_invalid_pathes=True
        )

    def _get_set_widget_value(self, widget, value=None):
        u"""widgetの値を取得または、値を設定.

        value=None で値を引き出す。それ以外では値を設定する

        widget ヴィジェットインスタンス
        value 設定する場合の値
        return 値設定時は None
               値取得時は list, int, bool, float, tuple, str等
        """
        # widgetのクラスを割り出す
        for basetype in self.BASETYPES_WRITE_OUT_WIDGET:
            if isinstance(widget, basetype):
                break

        # guiからの値取得
        if value is None:
            attr_name = self.ATTRNAME_GET_BY_WIDGET[basetype]
            return getattr(widget, attr_name)()

        # guiへのセット
        else:
            if isinstance(value, list):
                self._set_widget_list_value(widget, value)
                return
            attr_name = self.ATTRNAME_SET_BY_WIDGET[basetype]
            getattr(widget, attr_name)(value)

    def widget_dict_relative_to_json(self):
        u"""import, exportで扱うwidgetを収集.

        return dict(widget_name=widget)
        """
        widgets = dict()
        for name, widget in self.__dict__.items():
            if isinstance(widget, self.BASETYPES_WRITE_OUT_WIDGET):
                widgets[name] = widget
        return widgets

    def _import_setting(self):
        u"""セッティングをインポート."""
        # 引数設定
        kwargs = dict(
            caption='Import Setting File',
            filter='Json files (*.json)',
            options=QFileDialog.ReadOnly
        )
        # import, Exportで使用したディレクトリがあれば使いまわす
        if self.ever_handled_setting_path_exist is not None:
            kwargs['dir'] = self.ever_handled_setting_path_exist

        # josnファイルから設定を読み込み
        path = QFileDialog.getOpenFileName(self, **kwargs)[0]
        if not path:
            return
        self.ever_handled_setting_path_exist = os.path.dirname(path)

        # jsonの内容ををguiへ反映
        try:
            with open(path) as fobj:
                dic = json.load(fobj, object_pairs_hook=OrderedDict)
        except ValueError:
            msg = 'Invalid setting file:\n{}'.format(path)
            QMessageBox.critical(self, Setting.PACKAGE_NAME, msg)
            return

        # このツールで扱うjsonかをチェック
        if dic.get(Setting.TOOL_NAME_KEY, '') != Setting.PACKAGE_NAME:
            msg = 'Invalid setting file:\n{}'.format(path)
            QMessageBox.critical(self, Setting.PACKAGE_NAME, msg)
            return
        del dic[Setting.TOOL_NAME_KEY]

        # guiへ反映
        gui_dic = dic[Setting.GUI_KEY]
        for k, v in gui_dic.items():
            widget = getattr(self, k)
            self._get_set_widget_value(widget, v)

    def _export_setting(self):
        u"""セッティングをエクスポート."""
        # 引数設定
        kwargs = dict(
            caption='Export Setting File',
            filter='Json files (*.json)',
            options=QFileDialog.ReadOnly
        )
        # import, Exportで使用したディレクトリがあれば使いまわす
        if self.ever_handled_setting_path_exist is not None:
            kwargs['dir'] = self.ever_handled_setting_path_exist
        # 保存するファイルパス
        path = QFileDialog.getSaveFileName(self, **kwargs)[0]
        if not path:
            return
        self.ever_handled_setting_path_exist = os.path.dirname(path)
        # gui設定を保存
        self._save_gui_setting_to_json(path)

    def _save_gui_setting_to_json(self, path):
        u"""guiの設定を保存.

        一旦guiの設定をSettingへ送り、そのSettingの内容を保存
        path: unicode: 保存先パス
        return: [unicode,] レンダリングするシーンパスリスト
        """
        # Settingをgui内容でリフレッシュ
        self._send_setting()

        # Setting内容を収集
        if sys.version[0] == '2':
            str_typ = basestring
        else:
            str_typ = str
        setting_dict = dict()
        for k, v in Setting.__dict__.items():
            if not isinstance(v, (Number, str_typ, list, tuple, bool)):
                continue
            if k.startswith('__') or k.isupper():
                continue
            setting_dict[k] = v
        # Setting.scenes_pathesに関しては、ここで書き込み
        if not self.listwidget.item(0).text():
            setting_dict['scenes_pathes'] = []
        else:
            count = self.listwidget.count()
            setting_dict['scenes_pathes'] = [
                self.listwidget.item(i).text() for i in range(count)
            ]

        # guiの内容を取得
        gui_dict = dict()
        # 記録するwidgetを取得 dict(name=widget}
        widget_dict = self.widget_dict_relative_to_json()
        # widgetの各値を取得
        for name, widget in widget_dict.items():
            gui_dict[name] = self._get_set_widget_value(widget)

        # 書き込む辞書を作成
        record = dict()
        # このツールで扱うことを示す値を追加
        record[Setting.TOOL_NAME_KEY] = Setting.PACKAGE_NAME
        # setting, guiの要素を追加
        record[Setting.SETTING_KEY] = setting_dict
        record[Setting.GUI_KEY] = gui_dict
        # 保存
        with open(path, 'w') as fobj:
            json.dump(record, fobj, indent=2, sort_keys=True)

        return setting_dict['scenes_pathes']
