# -*-coding:utf-8-*-
u"""各DCCのメインとなる実行部.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

from .. import factory
from ..setting import Setting


class AbstractMainOperator(object):
    u"""実行メインの抽象クラス."""

    def _make_render_camera(self):
        u"""レンダリング用カメラカメラを作成.

        return: カメラオブジェクト
        """
        raise NotImplementedError

    def _get_specific_cameras(self, *camera_names):
        u"""指定済みの名前と一致するカメラを返す.

        camera_names: カメラ名イテレータ
        return: DCCツールのカメラ
        """
        raise NotImplementedError

    def _gather_all_meshes(self):
        u"""シーン内のメッシュオブジェクトを収集.

        return: メッシュリスト
        """
        raise NotImplementedError

    def _has_registered_mark(self, mesh_name):
        u"""キャプチャターゲットを制限するメッシュかを判定.

        判定用アトリビュートはbool型前提

        mesh_name: unicode
        return: bool
        """
        raise NotImplementedError

    def _extract_layer_visible(self, mesh_names):
        u"""アニメーション中、レイヤーによる非表示となるものを除去.

        mesh_names: [unicode,]: メッシュ名リスト
        return: [unicode,]: メッシュ名リスト
        """
        raise NotImplementedError

    def _get_point_container(self):
        u"""点群を収集.

        return DCCのPointContaier()
        """
        # メッシュ名の収集
        all_meshes = self._gather_all_meshes()
        if not Setting.capture_only_register_mesh:
            target_meshes = all_meshes
        else:
            extract_meshes = [
                x for x in all_meshes if self._has_registered_mark(x)
            ]
            # 登録メッシュが見当たらないなら全部のメッシュを対象とする
            if extract_meshes:
                target_meshes = extract_meshes
            else:
                target_meshes = all_meshes

        # レイヤーによる非表示設定を除外
        target_meshes = self._extract_layer_visible(target_meshes)

        # 点群のコンテナを用意
        point_container = factory.point_container()
        point_container.set_meshes(target_meshes)
        return point_container

    def execute(self):
        u"""レンダリング実行."""
        # シーンから得られる設定をSettingへ送る
        setter = factory.confirm_setting()
        setter.set_fps_by_current_scene()
        setter.set_coordinate_by_current_scene()
        if Setting.prefer_scene_range:
            setter.set_animation_range_by_current_scene()
        if Setting.prefer_scene_resolution:
            setter.set_resolution_by_current_scene()
        # レンダリングするカメラの取得または作成、新規カメラが必要なら点群取得
        if Setting.prefer_in_scene_camera:
            cameras = self._get_specific_cameras(*Setting.camera_filter_kwords)
            if not cameras:
                cameras = (self._make_render_camera(),)
                point_container = self._get_point_container()
            else:
                # 既存カメラがあるなら、点群を取得して計測する必要なし
                point_container = None
        else:
            # レンダリング用カメラ作成
            cameras = (self._make_render_camera(),)
            point_container = self._get_point_container()

        # 点群を包むカメラ操作インスタンスを用意
        fit_cameras = [factory.fit_camera(x) for x in cameras]

        # レンダリング
        renderer = factory.renderer(point_container, *fit_cameras)
        renderer.render()
