# -*-coding:utf-8-*-
u"""Blenderでののメインとなる実行部.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import math

import bpy
import mathutils

from ..setting import Setting
from . import main_func_abstract as mainabc


class MainOperator(mainabc.AbstractMainOperator):
    u"""実行メインの抽象クラス."""

    def _make_render_camera(self):
        u"""レンダリング用カメラカメラを作成.

        return: カメラオブジェクト
        """
        bpy.ops.object.camera_add(location=(0.0, 0.0, 0.0))
        camera = bpy.context.scene.objects.active
        # カメラの初期設定
        rot_xyz = (Setting.cam_rotx, Setting.cam_roty, Setting.cam_rotz)
        camera.rotation_euler = mathutils.Euler(
            map(math.radians, rot_xyz), 'XYZ'
        )
        # 焦点距離, clipping plane
        camera.data.lens = Setting.camera_focal_length
        camera.data.clip_start = Setting.camera_near
        camera.data.clip_end = Setting.camera_far
        return camera

    def _get_specific_cameras(self, *camera_names):
        u"""指定済みの名前と一致するカメラを返す.

        camera_names: カメラ名イテレータ
        return: [bpy.types.Object(),]: カメラリスト
        """
        class CameraName(object):
            u"""object名が指定名とマッチするかを調べる."""

            def __init__(self, obj):
                u"""consutructor.

                obj: bpy.types.Object()
                """
                self.obj = obj
                self.obj_name = obj.name

            def is_perfect_match(self):
                u"""指定名に完全一致するか."""
                return any(
                    x == self.obj_name for x in Setting.camera_filter_kwords
                )

            def is_partial_match(self):
                u"""指定名に部分一致するか."""
                return any(
                    x in self.obj_name for x in Setting.camera_filter_kwords
                )

        cams = [
            CameraName(x) for x in bpy.data.objects if
            isinstance(x.data, bpy.types.Camera)
        ]
        if Setting.camera_name_match_type == Setting.MatchType.kPerfectMatch:
            return [x.obj for x in cams if x.is_perfect_match()]
        else:
            return [x.obj for x in cams if x.is_partial_match()]

    def _gather_all_meshes(self):
        u"""シーン内のメッシュオブジェクトを収集.

        return: list(str,): メッシュオブジェクト名リスト
        """
        objs = bpy.context.scene.objects
        return [o.name for o in objs if isinstance(o.data, bpy.types.Mesh)]

    def _has_registered_mark(self, obj_name):
        u"""キャプチャターゲットを制限するメッシュかを判定.

        印はメッシュではなく、オブジェクトの方についている

        obj_name: unicode メッシュの、objectの方の名前
        return: bool
        """
        for obj in bpy.context.scene.objects:
            if Setting.REGISTERD_MARK in obj:
                if obj.name == obj_name:
                    return True
        else:
            return False

    def _extract_layer_visible(self, mesh_names):
        u"""アニメーション中、レイヤーによる非表示となるものを除去.

        mesh_names: [unicode,]: メッシュ名リスト
        return: [unicode,]: メッシュ名リスト
        """
        scene = bpy.context.scene
        selected_layers = [scene.layers[i] for i in range(20)]

        vis_meshes = set()

        # オブジェクトのレイヤと、シーンのレイヤが共にTrueであればvisible
        for obj in scene.objects:
            if not isinstance(obj.data, bpy.types.Mesh):
                continue
            obj_layers = [obj.layers[i] for i in range(20)]
            for i in range(20):
                if obj_layers[i] * selected_layers[i] == 1:  # True
                    vis_meshes.add(obj.data.name)
                    break

        return sorted(vis_meshes & set(mesh_names))


def execute_to_scene(scene_path):
    u"""シーンを開き、アニメーションにフィットするようにカメラを設定して、レンダリング.

    GUIを除けば、ここがメインファンクション

    scene_path: unicode: シーンのパス
    """
    if scene_path is None:
        bpy.ops.wm.read_homefile()
        return

    bpy.ops.wm.open_mainfile(filepath=scene_path)
    MainOperator().execute()
