# -*-coding:utf-8-*-
u"""各DCCのメインとなる実行部.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import maya.cmds as cmds
import maya.api.OpenMaya as om2

from ..setting import Setting
from . import main_func_abstract as mainabc


class MainOperator(mainabc.AbstractMainOperator):
    u"""実行メインの抽象クラス."""

    def _make_render_camera(self):
        u"""レンダリング用カメラカメラを作成.

        return: om2.MFnCamera()
        """
        new_cam = cmds.camera()[0]
        camera = om2.MFnCamera(om2.MSelectionList().add(new_cam).getDagPath(0))
        # カメラの初期設定
        rot_xyz = (Setting.cam_rotx, Setting.cam_roty, Setting.cam_rotz)
        cmds.xform(new_cam, ro=rot_xyz, ws=1, a=1)
        # 焦点距離, clipping plane
        cmds.setAttr('{}.fl'.format(new_cam), Setting.camera_focal_length)
        cmds.setAttr('{}.nearClipPlane'.format(new_cam), Setting.camera_near)
        cmds.setAttr('{}.farClipPlane'.format(new_cam), Setting.camera_far)
        return camera

    def _get_specific_cameras(self, *camera_names):
        u"""指定済みの名前と一致するカメラを返す.

        camera_names: カメラ名イテレータ
        return: [om2.MFnCamera(),]
        """
        class CameraName(object):
            u"""transform名が指定名とマッチするかを調べる."""

            def __init__(self, mcam):
                u"""constructor.

                mcam: om2.MFnCamera()
                """
                self.mcam = mcam
                self.trs_name = om2.MFnDagNode(mcam.parent(0)).name()

            def is_perfect_match(self):
                u"""指定名に完全一致するか."""
                return any(
                    x == self.trs_name for x in Setting.camera_filter_kwords
                )

            def is_partial_match(self):
                u"""指定名に部分一致するか."""
                return any(
                    x in self.trs_name for x in Setting.camera_filter_kwords
                )

        sel = om2.MSelectionList()
        for cam in cmds.ls(typ='camera'):
            sel.add(cam)
        mcams = [om2.MFnCamera(sel.getDagPath(i)) for i in range(sel.length())]
        cams = [CameraName(x) for x in mcams]

        if Setting.camera_name_match_type == Setting.MatchType.kPerfectMatch:
            return [x.mcam for x in cams if x.is_perfect_match()]
        else:
            return [x.mcam for x in cams if x.is_partial_match()]

    def _gather_all_meshes(self):
        u"""シーン内のメッシュオブジェクトを収集.

        return: list(unicode,): メッシュ名リスト
        """
        return cmds.ls(cmds.ls(dag=1, typ='mesh'), ni=1)

    def _has_registered_mark(self, mesh_name):
        u"""キャプチャターゲットを制限するメッシュかを判定.

        mesh_name: unicode
        return: bool
        """
        return cmds.attributeQuery(Setting.REGISTERD_MARK, n=mesh_name, ex=1)

    def _extract_layer_visible(self, mesh_names):
        u"""アニメーション中、レイヤーによる非表示となるものを除去.

        mesh_names: [unicode,]: メッシュ名リスト
        return: [unicode,]: メッシュ名リスト
        """
        hidden_state = {
            'visibility': False,
            'hideOnPlayback': True
        }
        context = om2.MDGContext()
        hidden_meshes = set()

        for mesh_name in mesh_names:
            if mesh_name in hidden_meshes:
                continue

            odag = om2.MGlobal.getSelectionListByName(mesh_name).getDagPath(0)

            # 親まで遡って、レイヤーによる不可視のメッシュを除去
            while not odag.hasFn(om2.MFn.kWorld):
                mdag = om2.MFnDagNode(odag)
                plug = mdag.findPlug('drawOverride', 0)
                src_plugs = plug.connectedTo(True, False)

                if not src_plugs:
                    odag = mdag.parent(0)
                    continue
                osrc = src_plugs[0].node()
                if not osrc.hasFn(om2.MFn.kDisplayLayer):
                    odag = mdag.parent(0)
                    continue

                # レイヤーのアトリビュートチェック
                mdg = om2.MFnDependencyNode(osrc)
                for attr, hidden in hidden_state.items():
                    vis = mdg.findPlug(attr, 0).asBool(context)
                    if vis == hidden:
                        break
                else:
                    odag = mdag.parent(0)
                    continue

                # レイヤー非表示のノード以下のメッシュを除外
                hiddens = set(cmds.ls(mdag.name(), dag=1, typ='mesh'))
                hidden_meshes |= hiddens
                odag = mdag.parent(0)

        return list(set(mesh_names) - hidden_meshes)


def execute_to_scene(scene_path=None):
    u"""シーンを開き、アニメーションにフィットするようにカメラを設定して、レンダリング.

    scene_path: unicode: シーンのパス
    """
    if scene_path is None:
        cmds.file(new=1, f=1)
        return

    cmds.file(scene_path, f=1, iv=1, o=1)
    MainOperator().execute()
