# -*-coding:utf-8-*-
u"""メッシュの情報頂点座標を取得.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
import bpy
import bmesh
import mathutils

from . import mesh_points_abstract as meshabc


class Mesh(meshabc.AbstractMesh):
    u"""バウンディングボックスを取得しやすくするためにbpy.types.Object()を管理."""

    def __init__(self, object_name):
        u"""construcetor."""
        self.obj = bpy.data.objects[object_name]

    @property
    def name(self):
        u"""メッシュの名前を返す."""
        return self.obj.name

    def _hidden_in_view(self):
        u"""ビューポート内で非表示状態にあるか."""
        return self.obj.hide

    def _current_points(self):
        u"""現在のフレームのポイントを返す.

        return: list of mathutils.Vector()
        """
        # bmeshに頂点のローカル座標を格納
        scene = bpy.context.scene
        obj_matrix = self.obj.matrix_world
        bm = bmesh.new()
        bm.from_object(self.obj, scene)
        locations = [obj_matrix * v.co for v in bm.verts]
        bm.free()
        return locations


class PointContainer(meshabc.AbstractPointContaier):
    u"""アニメーションを通した全体の点群を管理."""

    def clear(self):
        u"""self.pointsを空にする."""
        self.points = []
        self._center_position = None
        self._heights = None

    def set_meshes(self, object_names):
        u"""取得するメッシュのMeshインスタンスを認識させる."""
        self.meshes = [Mesh(n) for n in object_names]

    def get_current_points(self):
        u"""全メッシュの頂点を取得.

        return: list of mathutils.Vector()
        """
        points = []
        for mesh in self.meshes:
            points += mesh.get_current_points()
        if not len(points):
            points.append(mathutils.Vector((0.0, 0.0, 0.0)))
        return points
