# -*-coding:utf-8-*-
u"""メッシュの情報頂点座標を取得.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

import maya.api.OpenMaya as om2

from . import mesh_points_abstract as meshabc


class Mesh(meshabc.AbstractMesh):
    u"""バウンディングボックスを取得しやすくするためにom2.MFnMeshを管理."""

    def __init__(self, mesh_name):
        u"""construcetor."""
        sel = om2.MSelectionList().add(mesh_name)
        self.mesh = om2.MFnMesh(sel.getDagPath(0))

    @property
    def name(self):
        u"""メッシュのフルパス."""
        return self.mesh.fullPathName()

    def _hidden_in_view(self):
        u"""ビューポート内で非表示状態にあるか.

        階層を遡って調べる
        """
        current = self.mesh
        while True:
            plug = current.findPlug('v', 0)
            if plug.asBool(om2.MDGContext.kNormal) is False:
                return True
            parent = current.parent(0)
            if parent.hasFn(om2.MFn.kWorld):
                return False
            current = om2.MFnDagNode(parent)

    def _current_points(self):
        u"""現在のフレームのポイントを返す.

        return: om2.MPointArray()
        """
        return self.mesh.getPoints(om2.MSpace.kWorld)


class PointContainer(meshabc.AbstractPointContaier):
    u"""アニメーションを通した全体の点群を管理."""

    def clear(self):
        u"""self.pointsを空にする."""
        self.points = om2.MPointArray()
        self._center_position = None
        self._heights = None

    def set_meshes(self, mesh_names):
        u"""取得するメッシュのMeshインスタンスを認識させる."""
        self.meshes = [Mesh(x) for x in mesh_names]

    def get_current_points(self):
        u"""全メッシュの頂点を取得.

        return om2.MPointArray()
        """
        points = om2.MPointArray()
        for mesh in self.meshes:
            points += mesh.get_current_points()
        if not len(points):
            points.append(om2.MVector(0.0, 0.0, 0.0))  # 空は後の演算に悪影響
        return points
