# -*-coding:utf-8-*-
u"""各DCCの方法でハードウェアレンダーを実行する.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import os

from ..setting import Setting
from ..vector import calc_vector


class AbstractRenderer(object):
    u"""ハードウェアレンダーを行う."""

    def __init__(self, point_container, *fit_cameras):
        u"""constractor.

        fit_cameras: (FitCamera(), ): 任意の数
        point_container: PointContainer()
        """
        self.fit_cameras = fit_cameras
        self.point_container = point_container
        self._set_render_settings()

    def _set_render_settings(self):
        u"""DCCツールのレンダー設定を変更."""
        raise NotImplementedError

    def _set_current_time(self, frame):
        u"""指定フレームにカレントを設定."""
        raise NotImplementedError

    def _accumulate_points_whole_animation(self):
        u"""アニメーションにわたる点群を蓄積."""
        start = Setting.start_frame
        end = Setting.end_frame
        for frame in range(start, end + 1):
            self._set_current_time(frame)
            self.point_container.append_current_points()

    def _fit_camera(self):
        u"""カメラを点群にフィット."""
        points = self.point_container.points
        for camera in self.fit_cameras:
            camera.set_current_camera_position(points)

    def _fit_camera_to_current_frame(self):
        u"""カメラを現在のフレームの点群にフィット."""
        points = self.point_container.get_current_points()
        for camera in self.fit_cameras:
            camera.set_current_camera_position(points)

    def _set_keyframe_to_camera(self):
        u"""カメラにキーを打つ."""
        for camera in self.fit_cameras:
            camera.set_keyframe()

    def _set_keyframe_to_table(self):
        u"""ターンテーブルの階層を作成し、階層にキーを打つ."""
        if Setting.turntable_center == Setting.TurnCenter.kCenterGeometry:
            center2d = calc_vector.CalculateContactWithCircle.center2d
        else:
            center2d = (0.0, 0.0)
        for camera in self.fit_cameras:
            camera.set_turntable_animation(center2d)

    def _bake_camera_for_each_frame_figure(self):
        u"""毎フレーム、点群の再取得とカメラベイク."""
        for frame in range(Setting.start_frame, Setting.end_frame + 1):
            self._set_current_time(frame)
            self.point_container.clear()
            self._fit_camera_to_current_frame()
            self._set_keyframe_to_camera()

    def _bake_camera_for_whole_animation(self):
        u"""全フレームのアニメーション点群に対して、毎フレームカメラをフィット."""
        self._accumulate_points_whole_animation()
        for frame in range(Setting.start_frame, Setting.end_frame + 1):
            self._set_current_time(frame)
            self._fit_camera()
            self._set_keyframe_to_camera()

    def _stay_camera_for_whole_animation(self):
        u"""全フレームのアニメーション点群に対して、カメラを一か所にフィット."""
        self._accumulate_points_whole_animation()
        self._fit_camera()

    def _bake_camera_for_playblast(self):
        u"""カメラの動きを決める(プレイブラスト用)."""
        if Setting.camera_fit_mode == Setting.FitType.kEachFrameFigure:
            # フレーム毎に点群取得とカメラベイク
            self._bake_camera_for_each_frame_figure()
        elif Setting.camera_fit_mode == Setting.FitType.kWholeFigure:
            # 全フレームに渡る点群を取得してからカメラを全フレームベイク(カメラの回転アニメ反映)
            self._bake_camera_for_whole_animation()
        elif Setting.camera_fit_mode == Setting.FitType.kStay:
            # 全フレームに渡る点群を取得してから先頭フレームのみカメラにキーを打つ
            self._stay_camera_for_whole_animation()

    def _bake_camera_for_turntable(self):
        u"""カメラの動きを決める(ターンテーブル用)."""
        if Setting.camera_fit_mode_turntable == Setting.TurnTarget.kStay:
            self._set_current_time(Setting.start_frame)  # 最初の1フレームだけ取得
            self.point_container.append_current_points()
        elif Setting.camera_fit_mode_turntable == Setting.TurnTarget.kAnim:
            self._accumulate_points_whole_animation()
        self._fit_camera()
        self._set_keyframe_to_table()

    def _bake_camera(self):
        u"""カメラの動きを決める."""
        self.point_container.clear()
        if Setting.tool_purpose == Setting.ToolPurpose.kPlayblast:
            self._bake_camera_for_playblast()
        elif Setting.tool_purpose == Setting.ToolPurpose.kTurntable:
            self._bake_camera_for_turntable()

    def _set_background(self, camera_object):
        u"""カメラの背景色を設定.

        camera_object: FitCamera()
        """
        raise NotImplementedError

    def _render_each_camera(self, camera_object):
        u"""ハードウェアレンダーを実行.

        camera_object: FitCamera()
        """
        # path_maker = AbstractSavePath(camera_object.name)
        raise NotImplementedError

    def _set_viewport_shading(self):
        u"""ジオメトリの表示タイプを設定."""
        raise NotImplementedError

    def _set_ornament_invisible(self):
        u"""ロケータなどの表示切替え."""
        raise NotImplementedError

    def render(self):
        u"""ハードウェアレンダーを行う."""
        if self.point_container is not None:
            self._bake_camera()
        self._set_viewport_shading()
        if not Setting.show_ornament:
            self._set_ornament_invisible()
        for fit_camera in self.fit_cameras:
            self._set_background(fit_camera)
            self._render_each_camera(fit_camera)
            fit_camera.hide()


def join_path(*pathes):
    u"""パスを'/'で区切られた形に成形.

    pathes: *unicodes 連結させるパス
    return: 連結、成形済のパス
    """
    path = os.path.normpath(os.path.join(*pathes))
    path = path.replace(os.sep, '/')
    while True:
        new_path = path.replace('//', '/')
        if new_path == path:
            break
        path = new_path
    return new_path


def get_unique_folder(path):
    u"""フォルダ名の重複を避けるためにユニーク名にする.

    path: unicode: 変更前のフォルダパス
    return unicode: ユニークフォルダパス
    """
    new_path = path
    i = 0
    while True:
        if not os.path.isdir(new_path):
            break
        new_path = path + '_{:02}'.format(i)
        i += 1
    return new_path


class AbstractSavePath(object):
    u"""レンダーを保存するパスを決める."""

    def __init__(self, camera_name):
        u"""保存するパスを取得.

        camera_name: unicode: カメラ名
        """
        self._save_dir, self._file_name = self._get_unique_path_names(
            camera_name
        )

    @property
    def save_dir(self):
        u"""保存先フォルダパスを返す."""
        return self._save_dir

    @property
    def file_name(self):
        u"""保存ファイル名（拡張子抜き）を返す."""
        return self._file_name

    def _get_image_file_name(self, name):
        u"""ファイル名の重複を避けるためにユニーク名(画像用)を作成.

        path: unicode: 変更前のファイル名(拡張子抜き)
        return: unicode: ユニークファイル名(拡張子抜き)
        """
        return name

    def _get_unique_movie_file_name(self, path_name):
        u"""ファイル名の重複を避けるためにユニーク名を作成.

        path_name: unicode: 変更前のファイルパス(拡張子抜き)
        return: unicode: ユニークファイル名(拡張子抜き)
        """
        new_path_name = '{}_{:04}-{:04}'.format(
            path_name, Setting.start_frame, Setting.end_frame
        )
        i = 0
        while True:
            new_path = '{}.{}'.format(new_path_name, Setting.save_ext)
            if not os.path.isfile(new_path):
                break
            new_path_name = '{}_{:04}-{:04}_{:02}'.format(
                path_name, Setting.start_frame, Setting.end_frame, i
            )
            i += 1
        return os.path.basename(new_path_name)

    def _get_unique_path_names(self, camera_name):
        u"""保存フォルダ、ファイル名（拡張子抜き）をユニーク名で生成.

        camera_name: unicode: カメラ名
        return: tuple(unicode, unicode): 保存フォルダ名, ファイル名（拡張子抜き）
        """
        # 保存する場所を決めるルートディレクトリを取得
        if Setting.specific_save_root_dir:
            # 保存場所を絶対パスにする場合
            base_dir = Setting.specific_save_root_dir.replace(os.sep, '/')
            if base_dir.endswith('/'):
                base_dir = base_dir[:-1]
        else:
            # シーンの場所からの相対とする場合
            relative_dir = Setting.relative_save_dir.replace(os.sep, '/')
            if relative_dir.endswith('/'):
                relative_dir = relative_dir[:-1]
            base_dir = '{}/{}'.format(self._get_scene_dir(), relative_dir)
        # フォルダ、ファイルのユニーク名を取得
        scene_name = os.path.splitext(self._get_scene_basename())[0]
        file_name = '{}_{}'.format(scene_name, camera_name)  # 拡張子抜き
        if Setting.save_as_images:
            save_dir = get_unique_folder(
                '{}/{}'.format(base_dir, file_name)
            )
            # 拡張子抜きのファイル名
            file_name = self._get_image_file_name(file_name)
        else:
            save_dir = base_dir
            file_path = '{}/{}'.format(save_dir, file_name)
            # 拡張子抜きのファイル名
            file_name = self._get_unique_movie_file_name(file_path)
        return save_dir, file_name

    def _get_scene_path(self):
        u"""現在のシーンパスを返す."""
        raise NotImplementedError

    def _get_scene_dir(self):
        u"""現在のシーンのあるディレクトリを返す."""
        dirname = os.path.dirname(self._get_scene_path())
        return dirname.replace(os.sep, '/')

    def _get_scene_basename(self):
        u"""現在のシーン名を返す."""
        return os.path.basename(self._get_scene_path())
