# -*-coding:utf-8-*-
u"""blenderのハードウェアレンダーを実行する.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
import os
import sys

import bpy

from ..setting import Setting
from . import renderer_abstract as renabc


class AbstractSetRenderFormat:
    u"""Renderエリアの出力フォーマット設定を変更.

    Setting.qualityはblenderではcompression値として扱っている
    (Setting.quality=0 は compression=100 と同義)

    圧縮codecsはH.264固定
    """

    @classmethod
    def set_render_setting(cls):
        u"""Renderエリアの出力フォーマット設定を変更."""
        raise NotImplementedError

    @classmethod
    def _set_render_quality(cls):
        u"""圧縮設定."""
        ffmpeg = bpy.context.scene.render.ffmpeg
        ffmpeg.codec = 'H264'

        # 圧縮品質('LOSSLESS'はwindows10の標準プレイヤーで再生できなかったので除外)
        qualities = (
            'NONE', 'PERC_LOSSLESS', 'HIGH', 'MEDIUM',
            'LOW', 'VERYLOW', 'LOWEST'
        )
        span = int(100 / len(qualities))
        threasholds = range(0, 100, span)[:len(qualities)]
        quality_dic = {
            thres: quality for
            thres, quality in zip(threasholds, qualities)
        }
        # 圧縮品質へ代入する文字を、項目数を基に動的に決める
        for thres in reversed(sorted(quality_dic)):
            if Setting.quality > thres:
                ffmpeg.constant_rate_factor = quality_dic[thres]
                break


class Avi(AbstractSetRenderFormat):
    u"""拡張子avi."""

    @classmethod
    def set_render_settings(cls):
        u"""Renderエリアの出力フォーマット設定を変更."""
        # フォーマットタイプ
        image_settings = bpy.context.scene.render.image_settings
        if Setting.quality == 0:  # compression == 100
            image_settings.file_format = 'AVI_RAW'
        else:
            image_settings.file_format = 'FFMPEG'
            bpy.context.scene.render.ffmpeg.format = 'AVI'
            cls._set_render_quality()


class Mp4(AbstractSetRenderFormat):
    u"""拡張子mp4."""

    @classmethod
    def set_render_settings(cls):
        u"""Renderエリアの出力フォーマット設定を変更."""
        image_settings = bpy.context.scene.render.image_settings
        image_settings.file_format = 'FFMPEG'
        bpy.context.scene.render.ffmpeg.format = 'MPEG4'
        cls._set_render_quality()


class Png(AbstractSetRenderFormat):
    u"""拡張子png."""

    @classmethod
    def set_render_settings(cls):
        u"""Renderエリアの出力フォーマット設定を変更."""
        image_settings = bpy.context.scene.render.image_settings
        image_settings.file_format = 'PNG'
        bpy.context.scene.render.alpha_mode = 'TRANSPARENT'
        # Setting.quality=0 は compression=100 と同義
        image_settings.compression = Setting.quality


class Renderer(renabc.AbstractRenderer):
    u"""ハードウェアレンダーを行う."""

    def __init__(self, point_container, *fit_cameras):
        u"""constractor.

        fit_cameras: (FitCamera(), ): 任意の数
        point_container: PointContainer()
        """
        super(self.__class__, self).__init__(point_container, *fit_cameras)

        # ダミーコンテクストの作成、記述の簡略のため、アクティブなblenderUIオブジェクトを収集
        self.window = bpy.context.window_manager.windows[0]
        self.scene = bpy.context.scene
        self.blend_data = bpy.context.blend_data
        self.screen = self.window.screen
        for area in self.screen.areas:
            if area.type == 'VIEW_3D':
                self.area = area
                break
        self.space = self.area.spaces.active
        self.region = self.area.regions[-1]  # 3dビューでは必ず最後がビューポート

    def _set_render_settings(self):
        u"""レンダー設定.

        bpy.context.scene.render.image_settingsは _render
        """
        scene = bpy.context.scene
        render = scene.render
        render.resolution_x = Setting.resolution_x
        render.resolution_y = Setting.resolution_y
        render.resolution_percentage = Setting.resolution_percentage
        render.fps = Setting.fps
        render.use_file_extension = True
        scene.frame_start = Setting.start_frame
        scene.frame_end = Setting.end_frame
        scene.frame_step = 1

        # 保存フォーマット設定
        this_module = sys.modules[__name__]
        formatter = getattr(this_module, Setting.save_ext.title())
        formatter.set_render_settings()

    def _set_current_time(self, frame):
        u"""指定フレームにカレントを設定."""
        bpy.context.scene.frame_set(frame)

    def _set_background(self, _):
        u"""シーンの背景を変える."""
        if Setting.save_as_images:
            return  # blenderではimageのopengl()は強制で背景透過

        for theme in bpy.context.user_preferences.themes:
            gradients = theme.view_3d.space.gradients
            gradients.show_grad = False
            gradients.high_gradient = Setting.bg_color

    def _render_each_camera(self, fit_camera):
        u"""ハードウェアレンダーを実行.

        fit_camera: FitCamera()
        """
        # ハードウェアレンダーの保存先を変える
        path_maker = SavePath(fit_camera.name)
        save_dir = path_maker.save_dir
        file_name = path_maker.file_name

        if Setting.save_as_images:
            bpy.context.scene.render.filepath = '{}/{}'.format(
                save_dir, file_name
            )
        else:
            # Blenderでの動画は勝手にフレーム番号が入るので、一旦ダミーに。後で差し替え
            self.dummy_name_prefix = '{}_dummy_name_____'.format(
                Setting.PACKAGE_NAME
            )
            bpy.context.scene.render.filepath = '{}/{}'.format(
                save_dir, self.dummy_name_prefix
            )
        # ハードウェアレンダーのために、ビューをカメラ視点にする
        bpy.context.scene.camera = fit_camera.transform
        # カメラ視点にする
        self.space.region_3d.view_perspective = 'CAMERA'
        # ハードウェアレンダー実行
        context = {
            'scene': self.scene,
            'blend_data': self.blend_data,
            'window': self.window,
            'screen': self.screen,
            'area': self.area,
            'space_data': self.space,
            'region': self.region
        }
        try:
            bpy.ops.render.opengl(context, animation=1)
        except RuntimeError:
            import traceback
            raise Exception('{}{}{}'.format(
                bpy.data.filepath, os.linesep, traceback.format_exc()
            ))

        if Setting.save_as_images:
            return

        # 動画だった場合は、ダミー名で保存しているので、名前を差し替え
        for item in os.listdir(save_dir):
            if item.startswith(self.dummy_name_prefix):
                movie_path = '{}/{}.{}'.format(
                    save_dir, file_name, Setting.save_ext
                )
                os.rename('{}/{}'.format(save_dir, item), movie_path)
                break

    def _set_viewport_shading(self):
        u"""ジオメトリの表示タイプを設定."""
        # Cyclesはもとよりopengl()未対応
        if bpy.context.scene.render.engine == 'CYCLES':
            raise Exception(
                'Cycles not compatible. {}'.format(bpy.data.filepath)
            )

        # blender render, game engineならGLSL表示固定にする
        if bpy.context.scene.render.engine != 'CYCLES':
            bpy.context.scene.game_settings.material_mode = 'GLSL'  # 強制

        # Settingに応じてシェーディング表示変更
        shading_dic = {
            Setting.ShadingType.kWireframe: 'WIREFRAME',
            Setting.ShadingType.kShade: 'SOLID',
            Setting.ShadingType.kTexture: 'TEXTURED',
            Setting.ShadingType.kBeauty: 'MATERIAL'
        }

        shading_type = shading_dic[Setting.shading]
        if shading_type != 'TEXTURED':
            self.space.viewport_shade = shading_type
        # ビューポートをTEXTUREDにすとライティングが入るので、フラットシェードテクスチャにする
        else:
            self.space.viewport_shade = 'SOLID'
            self.space.show_textured_solid = True

    def _set_ornament_invisible(self):
        u"""ロケータなどの表示切替え."""
        self.space.show_only_render = True


class SavePath(renabc.AbstractSavePath):
    u"""レンダーを保存するパスを決める."""

    def __init__(self, camera_name):
        u"""保存するパスを取得.

        camera_name: unicode: カメラ名
        """
        super(SavePath, self).__init__(camera_name)

    def _get_image_file_name(self, path):
        u"""ファイル名の重複を避けるためにユニーク名(画像用)を作成.

        path: unicode: 変更前のファイル名(拡張子抜き)
        return: unicode: ユニークファイル名(拡張子抜き)
        """
        # blenderは画像ファイル名の最後に四桁のフレーム番号が付くので、拡張子との間に"_"を挿入
        return path + '_'

    def _get_scene_path(self):
        u"""現在のシーンパスを返す."""
        return bpy.data.filepath
