# -*-coding:utf-8-*-
u"""Mayaのハードウェアレンダーを実行する.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import inspect
import os

import maya.api.OpenMaya as om2
import maya.api.OpenMayaAnim as om2anm
import maya.cmds as cmds

from ..setting import Setting
from . import renderer_abstract as renabc


class PlayblastMovie(object):
    u"""playblastを動画で作成."""

    def __init__(self, save_dir, file_name, camera_name):
        u"""constractor.

        save_dir: unicode: 保存先フォルダ
        file_name: unicode: 保存ファイル名（拡張子抜き）
        camera_name: unicode: カメラ名
        """
        self.camera_name = camera_name
        self.kwargs = dict(
            format=Setting.save_ext,
            filename='{}/{}.{}'.format(save_dir, file_name, Setting.save_ext),
            sequenceTime=False,
            clearCache=True,
            viewer=False,
            showOrnaments=Setting.show_ornament,
            offScreen=True,
            framePadding=Setting.frame_padding,
            percent=100,
            compression=Setting.save_format,
            quality=Setting.quality,
            startTime=Setting.start_frame,
            endTime=Setting.end_frame,
            width=Setting.resolution_x,
            height=Setting.resolution_y
        )

    def do_playblast(self):
        u"""playblast実行."""
        # パネル選択
        current_panel = cmds.playblast(ae=1)
        cmds.setFocus(current_panel)
        # カメラのセットと実行
        cmds.lookThru(self.camera_name, current_panel)
        cmds.setAttr('{}.overscan'.format(self.camera_name), 1.0)
        cmds.playblast(**self.kwargs)


class PlayblastImages(PlayblastMovie):
    u"""playblastを静止画で作成."""

    def __init__(self, save_dir, file_name, camera_name):
        u"""constractor.

        save_dir: unicode: 保存先フォルダ
        file_name: unicode: 保存ファイル名（拡張子抜き）
        camera_name: unicode: カメラ名
        """
        super(PlayblastImages, self).__init__(save_dir, file_name, camera_name)
        self.kwargs.update(dict(
            format='image',
            compression=Setting.save_ext,
        ))


# fpsのDCC記述
FPS_DICT = {
    15: 'game',
    24: 'film',
    25: 'pal',
    30: 'ntsc',
    48: 'show',
    50: 'palf',
    60: 'ntscf'
}


class Renderer(renabc.AbstractRenderer):
    u"""ハードウェアレンダーを行う."""

    def _set_render_settings(self):
        u"""レンダー設定（Mayaはプレイブラスト部で多くを設定する）."""
        cmds.currentUnit(t=FPS_DICT[Setting.fps])

    def _set_current_time(self, frame):
        u"""指定フレームにカレントを設定."""
        om2anm.MAnimControl.setCurrentTime(om2.MTime(frame, om2.MTime().unit))

    def _set_background(self, fit_camera):
        u"""撮影カメラのみにバックグラウンドイメージを反映させる.

        fit_camera: FitCamera()
        """
        for img in cmds.ls(typ='imagePlane'):
            cmds.setAttr('{}.v'.format(img), 0)

        # 画像保存で背景を透明にするならimagePlaneを貼らない
        if Setting.bg_mask and Setting.save_ext not in ('avi', 'mp4', 'mov'):
            return

        # 新規プレーン
        current = inspect.stack()[0][1]
        try:
            image_path = os.path.normpath(os.path.join(
                current, '../../../resources/background_white.png'
            ))
        except UnicodeDecodeError:  # このパッケージに日本語パスが含まれる場合
            image_path = os.path.normpath(os.path.join(
                current, b'../../../resources/background_white.png'
            ))
        image = cmds.imagePlane(c=fit_camera.name, fn=image_path)[1]
        cmds.setAttr(image + '.depth', Setting.camera_far)

        # プレーンのサイズ調整のため、レンダリング解像度をシーンへ適用
        res = 'defaultResolution.'
        pixel_aspect = cmds.getAttr(res + 'pixelAspect')
        cmds.setAttr(res + 'width', Setting.resolution_x)
        cmds.setAttr(res + 'height', Setting.resolution_y)
        device_aspect = float(Setting.resolution_x) / Setting.resolution_y
        cmds.setAttr(res + 'deviceAspectRatio', device_aspect)
        cmds.setAttr(res + 'pixelAspect', pixel_aspect)

        # プレーンをResolution Gateへフィット
        if Setting.resolution_y > Setting.resolution_x:
            cmds.setAttr(image + '.fit', 4)  # ToSize
            inverse_ratio = float(Setting.resolution_y) / Setting.resolution_x
            plane_x = cmds.getAttr(image + '.sizeX')
            cmds.setAttr(image + '.sizeY', plane_x * inverse_ratio)
        else:
            cmds.setAttr(image + '.fit', 2)  # Horizontal

        # 色
        cmds.setAttr(
            '{}.colorGain'.format(image), *Setting.bg_color, typ='double3'
        )

        # イメージのレンダリングミスを防ぐため、clipping planeを伸ばす
        cmds.setAttr(
            fit_camera.name + '.farClipPlane', Setting.camera_far * 1.5
        )

    def _render_each_camera(self, fit_camera):
        u"""ハードウェアレンダーを実行.

        fit_camera: FitCamera()
        """
        # arnordの背景バグを回避するためにHardwareRender2.0へ
        cmds.setAttr(
            'defaultRenderGlobals.ren', 'mayaHardware2', type='string'
        )

        # 保存パス
        path_maker = SavePath(fit_camera.transform.name())
        save_dir = path_maker.save_dir
        file_name = path_maker.file_name
        args = (save_dir, file_name, fit_camera.name)

        if Setting.save_as_images:
            playblaster_cls = PlayblastImages
        else:
            playblaster_cls = PlayblastMovie

        playblaster = playblaster_cls(*args)
        playblaster.do_playblast()

    def _set_viewport_shading(self):
        u"""ジオメトリの表示タイプを設定."""
        pnl = cmds.playblast(ae=1)
        if Setting.shading == Setting.ShadingType.kWireframe:
            cmds.modelEditor(
                pnl, e=1, da='wireframe'
            )
        elif Setting.shading == Setting.ShadingType.kShade:
            cmds.modelEditor(
                pnl, e=1, da='smoothShaded', dtx=0, dl='default'
            )
        elif Setting.shading == Setting.ShadingType.kTexture:
            cmds.modelEditor(
                pnl, e=1, da='smoothShaded', dtx=1, dl='default'
            )
        elif Setting.shading == Setting.ShadingType.kBeauty:
            cmds.modelEditor(
                pnl, e=1, da='smoothShaded', dtx=1, dl='all'
            )
        else:
            raise Exception

    def _set_ornament_invisible(self):
        u"""ロケータなどの表示切替え."""
        current_panel = cmds.playblast(ae=1)
        cmds.modelEditor(current_panel, e=1, alo=0, pm=1)
        cmds.modelEditor(current_panel, e=1, imp=1)


class SavePath(renabc.AbstractSavePath):
    u"""レンダーを保存するパスを決める."""

    def __init__(self, camera_name):
        u"""保存するパスを取得.

        camera_name: unicode: カメラ名
        """
        super(SavePath, self).__init__(camera_name)

    def _get_scene_path(self):
        u"""現在のシーンパスを返す."""
        return cmds.file(q=1, sn=1)
