# -*-coding:utf-8-*-
u"""セッティング.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import tempfile


class Setting(object):
    u"""各種セッティング."""

    HELP_URL = 'http://maya2blender.xsrv.jp/tools/' \
               'hardware_render_animations/v020/document/index.html'

    MAYA2018 = 'maya2018'
    BLENDER279B = 'blender279b'
    BLENDER280BETA = 'blender280beta'

    PACKAGE_NAME = 'hardware_render_animations'  # 相対インポートで使用
    REGISTERD_MARK = 'marked_by_hardware_render_animations'  # メッシュ限定の印

    LOG_PATH = '{}/errorlog_{}.txt'.format(tempfile.gettempdir(), PACKAGE_NAME)

    # jsonで設定をimport, exportする際のキー
    # {
    #   '_tool_name': ツール名,
    #   'gui': {widget名: 値},
    #   'setting': {Settingの項目: 値},
    # }
    # の形式でjsonへ保存される
    TOOL_NAME_KEY = '_tool_name'
    GUI_KEY = 'gui'
    SETTING_KEY = 'setting'

    # レンダーをサブプロセスで起動したDCCから行う場合に使用する設定保存パス
    TEMPFILE_PATH = '{}/{}_setting.json'.format(
        tempfile.gettempdir(), PACKAGE_NAME
    )

    class ToolPurpose(object):
        u"""ハードウェアレンダーとターンテーブルのどちらのツールとして使用するか."""

        kPlayblast = 0
        kTurntable = 1

    class FitType(object):
        u"""プレイブラストでのカメラのフィッティング対象."""

        kWholeFigure = 0  # アニメーションを通したオブジェクトの変形全体にフィット(重い)
        kEachFrameFigure = 1  # フレーム毎にオブジェクトにフィット
        kStay = 2  # カメラは一度決めたら動かない。フィット対象はアニメーション全体

    class TurnTarget(object):
        u"""ターンテーブルの対象ジオメトリ."""

        kAnim = 0  # アニメーションを通したオブジェクトの変形全体を捉える
        kStay = 1  # 最初のフレームの形状のみとらえる

    class TurnCenter(object):
        u"""ターンテーブルの回転中心."""

        kOrigin = 0  # 原点
        kCenterGeometry = 1  # ジオメトリの中心

    class ShadingType(object):
        u"""ビューポートシェーディング."""

        kWireframe = 0
        kShade = 1
        kTexture = 2
        kBeauty = 3

    class MatchType(object):
        u"""単語のマッチ検索をするタイプ."""

        kPerfectMatch = 0  # 完全一致
        kPartialMatch = 1  # 部分一致

    caller_dcc_path = None  # 別プロセスから呼ばれた場合に、呼んだDCCツールのパスを記録

    tool_purpose = ToolPurpose.kTurntable
    dcc_tool = MAYA2018  # ツールで使用するモジュールを選択するためのキー
    coord_up_id = 1  # 座標tupleの、上空にあたるインデックス
    coord_gnd_id = 2  # 座標tupleの(x以外の)地面にあたるインデックス

    resolution_x = 1280  # レンダリング解像度(横)
    resolution_y = 720  # レンダリング解像度(縦)
    resolution_percentage = 100
    start_frame = 0  # int
    end_frame = 30  # int
    fps = 30
    camera_fit_mode = FitType.kStay  # FitTypeのアトリビュートのどれか
    camera_focal_length = 80.0
    camera_near = 1.0  # クリッピングプレーン
    camera_far = 100.0
    cam_rotx = 0.0  # カメラを新規作成する場合のrotation(degrees)
    cam_roty = 0.0
    cam_rotz = 0.0
    count_hide_vertices = False  # 非表示になっている部分をレンダー範囲に考慮するか
    render_margin_percent = 0  # レンダー領域の外側のマージンパーセント
    camera_rotate_change = False  # カメラの回転をアニメーション中に再計算するか

    turntable_center = TurnCenter.kOrigin  # TurnCenterのattr
    camera_fit_mode_turntable = TurnTarget.kAnim  # TurnTargetのattr
    turntimes_during_animation = 2

    save_as_images = False  # Falseならファイルが、Trueならフォルダが作られる
    save_ext = 'avi'  # 'avi'  # 保存ファイルの拡張子
    save_format = 'AVI_JPEG'  # 保存フォーマットなど
    quality = 50  # レンダー品質
    specific_save_root_dir = ''  # 保存するルートフォルダ（''ならシーンと同じところ）
    relative_save_dir = ''  # specific_save_root_dirからの相対パス
    show_ornament = False  # ロケータなどを非表示にするか（可能なら）
    frame_padding = 4
    prefer_scene_range = False  # start, endフレームをシーンから取得するか
    prefer_scene_resolution = False  # シーンの出力解像度を優先するか
    bg_color = (0.5, 0.5, 0.5)  # 背景色
    bg_mask = True

    capture_only_register_mesh = False  # 別途指定したメッシュのみを対象にするか

    shading = ShadingType.kShade  # キャプチャ時のビューポートシェーディング
    prefer_in_scene_camera = False  # シーンにあるカメラを優先するか
    camera_filter_kwords = tuple()  # シーンカメラを使う際の、カメラ名フィルタワード
    camera_name_match_type = MatchType.kPartialMatch  # カメラ名の一致タイプ

    scenes_pathes = tuple()  # レンダリング対象となるシーン

    @classmethod
    def scene_exts(cls):
        u"""DCCツールで使用している拡張子を返す.

        return: tuple(unicode)
        """
        exts = {
            cls.MAYA2018: ('.ma', '.mb'),
            cls.BLENDER279B: ('.blend',),
        }
        return exts[cls.dcc_tool]

    @classmethod
    def set_by_dict(cls, setting_dict):
        u"""辞書を基に、このクラスのアトリビュートを設定.

        setting_dict: dict(): e.g. {'resolution_x': 1920, ...}
        """
        for k, v in setting_dict.items():
            setattr(cls, k, v)
