# -*-coding:utf-8-*-
u"""Vector3Dを演算する.

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

Author: Masayuki Osaka
"""
from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function
import math

from ..setting import Setting


def get_cross_line_of_2planes(p1, p2, n1, n2):
    u"""二平面の交線を求める.

    p1: Vector3D(): 平面1上の点
    p2: Vector3D(): 平面2上の点
    n1: Vector3D(): 平面1の法線
    n2: Vector3D(): 平面2の法線
    return tuple(Vector3D(), Vector3D()): 交線の点, 交線のベクトル
    """
    nx = n1 ^ n2
    nt = nx ^ n1  # 平面1から交線への垂線ベクトル
    t = (n2 * (p2 - p1)) / (n2 * nt)
    px = t * nt + p1
    return px, nx


def get_closest_points_of_2lines(p1, p2, n1, n2):
    u"""二直線の最近接点を求める.

    p1: Vector3D(): 直線1上の点
    p2: Vector3D(): 直線2上の点
    n1: Vector3D(): 直線1のベクトル
    n2: Vector3D(): 直線2のベクトル
    return tuple(Vector3D(), Vector3D()):
           直線2への直線1上の最近接点, 直線1への直上2上の最近接点
    """
    n1 = n1.normalize()
    n2 = n2.normalize()
    n1n2 = n1 * n2
    n2pp = n2 * (p2 - p1)
    d1 = (n1 * (p2 - p1) - n2pp * n1n2) / (1 - n1n2 * n1n2)
    d2 = d1 * n1n2 - n2pp
    return p1 + d1 * n1, p2 + d2 * n2


def get_normalized_2d(x, y):
    u"""二次元ベクトルの正規化.

    x: float
    y: float
    return: tuple(float, float): 正規化ベクトル
    """
    coef = 1.0 / math.sqrt((x ** 2) + (y ** 2))
    return x * coef, y * coef


class CalculateContactWithCircle(object):
    u"""各点が描く円群に対し、ベクトルが接する点を取得.

    ベクトルの始点が三次元空間上にあり、かつ方向が円弧の中心を向く位置の、始点と円の接点を計算
    """

    @classmethod
    def initialize_cache(cls):
        u"""クラス変数の値を変える前に初期化すること."""
        cls.center2d = None  # tuple(float, float): 点群の二次元中心
        cls.points2d = None  # list(tuple(float, float),): 点群を二次元正規ベクトル
        cls.heights = None  # list(float,): 各点の元々の高さ
        cls.distances2d = None  # list(float,): 二次元での点群と中心点の距離(半径)

    @classmethod
    def get_contacts(cls, vector, points):
        u"""接点取得.

        vector: 各点が描く円群に対し接したい、各DCCツールのベクトル
        points: 各DCCツールの点群
        return: list(tuple(float, float, float),): 円弧とベクトル始点の接点
                (ベクトルの始点が三次元空間上にあり、かつ方向が円弧の中心を向く位置として)
        """
        # 初期設定
        cls.vector3d = vector
        cls.points3d = points
        # 呼出しにかかわらず変化しない値を設定
        cls._set_points2d()
        cls._set_center2d()
        cls._set_heights()
        cls._set_distances2d()

        # 点群との接点を計算
        norm_vec2d = get_normalized_2d(vector[0], vector[Setting.coord_gnd_id])
        minus_x2d = norm_vec2d[0] * -1
        minus_y2d = norm_vec2d[1] * -1
        if Setting.coord_up_id == 1:  # Mayaなど
            contacts = [
                (
                 minus_x2d * cls.distances2d[i] + cls.center2d[0],
                 cls.heights[i],
                 minus_y2d * cls.distances2d[i] + cls.center2d[1]
                ) for i in range(len(cls.points3d))
            ]
        else:  # blenderなど
            contacts = [
                (
                 minus_x2d * cls.distances2d[i] + cls.center2d[0],
                 minus_y2d * cls.distances2d[i] + cls.center2d[1],
                 cls.heights[i]
                ) for i in range(len(cls.points3d))
            ]
        return contacts

    @classmethod
    def _set_points2d(cls):
        u"""二次元投影した点群を作成."""
        if cls.points2d is not None:
            return
        # 正規化前の、平面投影ベクトルを作成
        cls.points2d = [(p[0], p[Setting.coord_gnd_id]) for p in cls.points3d]

    @classmethod
    def _set_center2d(cls):
        u"""二次元投影した点群の平均を作成."""
        if cls.center2d is not None:
            return
        if Setting.turntable_center == Setting.TurnCenter.kOrigin:
            cls.center2d = (0.0, 0.0, 0.0)
            return
        length = len(cls.points3d)
        gnd_id = Setting.coord_gnd_id
        average_x2d = sum(p[0] for p in cls.points3d) / length
        average_y2d = sum(p[gnd_id] for p in cls.points3d) / length
        cls.center2d = (average_x2d, average_y2d)

    @classmethod
    def _set_heights(cls):
        u"""点群の高さを作成."""
        if cls.heights is not None:
            return
        up_id = Setting.coord_up_id
        cls.heights = [p[up_id] for p in cls.points3d]

    @classmethod
    def _set_distances2d(cls):
        u"""二次元点群と二次元中心点から描かれる円の、半径を作成."""
        if cls.distances2d is not None:
            return
        c = cls.center2d
        cls.distances2d = [
            math.sqrt(((p[0] - c[0]) ** 2) + ((p[1] - c[1]) ** 2)) for
            p in cls.points2d
        ]
